import six

from collections import defaultdict
import itertools
import json
import os
import posixpath
import re
import requests

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from ....lib import red, toiter, Bbox, Vec, jsonify
from ....mesh import Mesh
from .... import paths
from ....scheduler import schedule_jobs

from ...precomputed.mesh import UnshardedLegacyPrecomputedMeshSource, PrecomputedMeshMetadata


class GrapheneUnshardedMeshSource(UnshardedLegacyPrecomputedMeshSource):

  def compute_filename(self, label):
    layer_id = self.meta.meta.decode_layer_id(label)
    chunk_block_shape = Vec(*self.meta.meta.mesh_chunk_size, dtype=np.int64)
    chunk_block_shape *= np.int64(self.meta.meta.fan_out ** max(0, layer_id - 2))
    start = self.meta.meta.decode_chunk_position(label)
    start *= chunk_block_shape
    bbx = Bbox(start, start + chunk_block_shape)
    return "{}:0:{}".format(label, bbx.to_filename())

  def exists(self, labels, progress=None):
    """
    Checks for dynamic mesh existence.
  
    Returns: { label: boolean, ... }
    """
    labels = toiter(labels)
    filenames = [
      self.compute_filename(label) for label in labels
    ]

    cloudpath = self.meta.join(self.meta.cloudpath, self.meta.mesh_path)
    return CloudFiles(cloudpath, secrets=self.config.secrets).exists(filenames)

  def get_fragment_labels(self, segid, lod=0, level=2, bbox=None, bypass=False):
    if bypass:
      return [ segid ]

    manifest = self.fetch_manifest(segid, lod, level, bbox, return_segids=True, verify=False)
    return manifest["seg_ids"]

  def get_fragment_filenames(self, segid, lod=0, level=2, bbox=None, bypass=False):
    if bypass:
      return [ self.compute_filename(segid) ]

    manifest = self.fetch_manifest(segid, lod, level, bbox, verify=True)
    return manifest["fragments"]

  def fetch_manifest(self, segid, lod=0, level=2, bbox=None, return_segids=False, verify=True):
    """
    verify: the server calls exists and returns byte ranges. this is intended 
      to take advantage of the lower latency on the server side, but can be a
      bottleneck.
    """
    # TODO: add lod to endpoint
    query_d = {
      'verify': bool(verify),
    }
    if return_segids:
      query_d['return_seg_ids'] = 1

    if bbox is not None:
      bbox = Bbox.create(bbox)
      query_d['bounds'] = bbox.to_filename()

    url = "%s/%s:%s" % (self.meta.meta.manifest_endpoint, segid, lod)
    if level is not None:
      res = requests.get(
        url,
        data=jsonify({ "start_layer": level }),
        params=query_d,
        headers=self.meta.meta.auth_header
      )
    else:
      res = requests.get(url, params=query_d, headers=self.meta.meta.auth_header)

    res.raise_for_status()

    return json.loads(res.content.decode('utf8'))

  def download_segid(self, seg_id, bounding_box, bypass, use_byte_offsets=True):
    """
    Download a mesh for a single segment ID.

    seg_id: Download the mesh for this segid.
    bounding_box: Limit the query for child meshes to this bounding box.
    bypass: Don't fetch the manifest, precompute the filename instead. Use this
      only when you know the actual mesh labels in advance.
    use_byte_offsets: Applicable only for the sharded format. Reuse the byte_offsets
      into the sharded format that the server precalculated to accelerate download.
      A time when you might want to switch this off is when you're working on a new
      meshing job with different sharding parameters but are keeping the existing 
      meshes for visualization while it runs.
    allow_missing: If set to True, return None if segid missing. If set to False, throw
      an error.
    """
    import DracoPy

    level = self.meta.meta.decode_layer_id(seg_id)
    fragment_filenames = self.get_fragment_filenames(
      seg_id, level=level, bbox=bounding_box, bypass=bypass
    )
    fragments = self._get_mesh_fragments(fragment_filenames)
    fragments = sorted(fragments, key=lambda frag: frag[0])  # make decoding deterministic

    fragiter = tqdm(fragments, disable=(not self.config.progress), desc="Decoding Mesh Buffer")
    is_draco = False
    for i, (filename, frag) in enumerate(fragiter):
      mesh = None
      
      if frag is not None:
        try:
          # Easier to ask forgiveness than permission
          mesh = Mesh.from_draco(frag)
          is_draco = True
        except DracoPy.FileTypeException:
          mesh = Mesh.from_precomputed(frag)
          
      fragments[i] = mesh
    
    fragments = [ f for f in fragments if f is not None ] 
    if len(fragments) == 0:
      raise IndexError('No mesh fragments found for segment {}'.format(seg_id))

    mesh = Mesh.concatenate(*fragments)
    mesh.segid = seg_id
    return mesh, is_draco

  def get(
      self, segids, 
      remove_duplicate_vertices=False, 
      fuse=False, bounding_box=None,
      bypass=False, use_byte_offsets=False,
      deduplicate_chunk_boundaries=True,
      allow_missing=False,
    ):
    """
    Merge fragments derived from these segids into a single vertex and face list.

    Why merge multiple segids into one mesh? For example, if you have a set of
    segids that belong to the same neuron.

    segid: (iterable or int) segids to render into a single mesh

    Optional:
      remove_duplicate_vertices: bool, fuse exactly matching vertices within a chunk
      fuse: bool, merge all downloaded meshes into a single mesh
      bounding_box: Bbox, bounding box to restrict mesh download to
      bypass: bypass requesting the manifest and attempt to get the 
        segids from storage directly by testing the dynamic and then the initial mesh. 
        This is an exceptional usage of this tool and should be applied only with 
        an understanding of what that entails.
      use_byte_offsets: For sharded volumes, we can use the output of 
        exists(..., return_byte_offsets) that the server already did in order
        to skip having to query the sharded format again.
      deduplicate_chunk_boundaries: Our meshing is done in chunks and creates duplicate vertices
        at the boundaries of chunks. This parameter will automatically deduplicate these if set
        to True. Superceded by remove_duplicate_vertices.
      allow_missing: If set to True, missing segids will be ignored. If set to False, an error
        is thrown.
    
    Returns: Mesh object if fused, else { segid: Mesh, ... }
    """
    segids = list(set([ int(segid) for segid in toiter(segids) ]))
    meta = self.meta.meta

    meshes = []
    for seg_id in tqdm(segids, disable=(not self.config.progress), desc="Downloading Meshes"):
      level = meta.decode_layer_id(seg_id)
      if allow_missing:
        try:
          mesh, is_draco = self.download_segid(
            seg_id, bounding_box, bypass, use_byte_offsets
          )
        except IndexError:
          continue
      else:
        mesh, is_draco = self.download_segid(
          seg_id, bounding_box, bypass, use_byte_offsets
        )
      resolution = meta.resolution(self.config.mip)
      if meta.chunks_start_at_voxel_offset:
        offset = meta.voxel_offset(self.config.mip)
      else:
        offset = Vec(0,0,0)

      if remove_duplicate_vertices:
        mesh = mesh.consolidate()
      elif is_draco:
        if not deduplicate_chunk_boundaries:
          pass
        elif level == 2:
          # Deduplicate at quantized lvl2 chunk borders
          draco_grid_size = meta.get_draco_grid_size(level)
          mesh = mesh.deduplicate_chunk_boundaries(
            meta.mesh_chunk_size * resolution,
            offset=offset * resolution,
            is_draco=True,
            draco_grid_size=draco_grid_size,
          )
        else:
          # TODO: cyclic draco quantization to properly
          # stitch and deduplicate draco meshes at variable
          # levels (see github issue #299)
          print('Warning: deduplication not currently supported for this layer\'s variable layered draco meshes')
      elif deduplicate_chunk_boundaries:
        mesh = mesh.deduplicate_chunk_boundaries(
            meta.mesh_chunk_size * resolution,
            offset=offset * resolution,
            is_draco=False,
          )
      
      meshes.append(mesh)

    if not fuse:
      return { m.segid: m for m in meshes }

    return Mesh.concatenate(*meshes).consolidate()




import pytest 
import sys

import cloudvolume

# Basic test of sharded meshes
# The following test files are used from https://storage.googleapis.com:
#   /neuroglancer-janelia-flyem-hemibrain/v1.0/segmentation/info
#   /fafb-ffn1-20190805/segmentation/mesh/info
#   /neuroglancer-janelia-flyem-hemibrain/v1.0/segmentation/mesh/171.shard
#   /neuroglancer-janelia-flyem-hemibrain/v1.0/segmentation/mesh/193.shard
#   /neuroglancer-janelia-flyem-hemibrain/v1.0/segmentation/mesh/185.shard 

@pytest.fixture
def hemibrain_vol():
  return cloudvolume.CloudVolume('gs://neuroglancer-janelia-flyem-hemibrain/v1.0/segmentation', mip=0, cache=False, use_https=True)

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3")
def test_get_sharded_mesh(hemibrain_vol):
  exists = hemibrain_vol.mesh.exists([511271574, 360284300])
  assert(all(exists))

  exists = hemibrain_vol.mesh.exists([666666666, 666666667, 666666668])
  assert(all([a == None for a in exists]))

  meshes = hemibrain_vol.mesh.get([511271574, 360284300], lod=2)
  assert len(meshes[511271574].faces) == 258647

  meshes = hemibrain_vol.mesh.get([511271574, 360284300], lod=3)
  assert len(meshes[511271574].faces) == 50501

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3")
def test_get_sharded_mesh_invalid_lod(hemibrain_vol):
  try:
    meshes = hemibrain_vol.mesh.get([511271574, 360284300], lod=8)
  except ValueError:
    pass
  except:
    raise



def content_type(encoding):
  if encoding == 'jpeg':
    return 'image/jpeg'
  elif encoding in ('compressed_segmentation', 'fpzip', 'kempressed'):
    return 'image/x.' + encoding 
  return 'application/octet-stream'

def should_compress(encoding, compress, cache, iscache=False):
  if iscache and cache.compress != None:
    return cache.compress

  if compress is None:
    return 'gzip' if encoding in ('raw', 'compressed_segmentation') else None
  elif compress == True:
    return 'gzip'
  elif compress == False:
    return None
  else:
    return compress

def cdn_cache_control(val):
  """Translate cdn_cache into a Cache-Control HTTP header."""
  if val is None:
    return 'max-age=3600, s-max-age=3600'
  elif type(val) is str:
    return val
  elif type(val) is bool:
    if val:
      return 'max-age=3600, s-max-age=3600'
    else:
      return 'no-cache'
  elif type(val) is int:
    if val < 0:
      raise ValueError(
        'cdn_cache must be a positive integer, boolean, or string. Got: ' + str(val)
      )

    if val == 0:
      return 'no-cache'
    else:
      return 'max-age={}, s-max-age={}'.format(val, val)
  else:
    raise NotImplementedError(type(val) + ' is not a supported cache_control setting.')



from ..precomputed.sharding import ShardingSpecification, ShardReader

class GrapheneShardReader(ShardReader):
  def compute_shard_location(self, label):
    shard_loc = self.spec.compute_shard_location(label)
    chunk_pos_num = self.meta.meta.decode_chunk_position_number(label)
    filename = str(chunk_pos_num) + '-' + str(shard_loc.shard_number) + '.shard'
    return (filename, shard_loc.minishard_number)


 


from .mesh import GrapheneMeshSource
from .metadata import GrapheneMetadata
from ..precomputed.image import PrecomputedImageSource
from ..precomputed.skeleton import PrecomputedSkeletonSource

from ...cacheservice import CacheService
from ...cloudvolume import SharedConfiguration, register_plugin
from ...paths import strict_extract

from requests import HTTPError

def create_graphene(
    cloudpath, mip=0, bounded=True, autocrop=False,
    fill_missing=False, cache=False, compress_cache=None,
    cdn_cache=True, progress=False, info=None, provenance=None,
    compress=None, parallel=1,
    delete_black_uploads=False, background_color=0,
    green_threads=False, use_https=False,
    mesh_dir=None, skel_dir=None, agglomerate=False, 
    secrets=None, **kwargs
  ):
    from ...frontends import CloudVolumeGraphene
    
    path = strict_extract(cloudpath)
    config = SharedConfiguration(
      cdn_cache=cdn_cache,
      compress=compress,
      compress_level=None,
      green=green_threads,
      mip=mip,
      parallel=parallel,
      progress=progress,
      secrets=secrets,
    )

    def mkcache(cloudpath):
      return CacheService(
        cloudpath=(cache if type(cache) == str else cloudpath),
        enabled=bool(cache),
        config=config,
        compress=compress_cache,
      )
    meta = GrapheneMetadata(
      cloudpath, config=config, cache=mkcache(cloudpath),
      info=info, provenance=provenance, 
      use_https=use_https, agglomerate=agglomerate,
      auth_token=config.secrets,
    )
    # Resetting the cache is necessary because
    # graphene retrieves a data_dir from the info file
    # that reflects the real cache location.
    cache = mkcache(meta.cloudpath) 
    meta.cache = cache
    cache.meta = meta

    if mesh_dir:
      meta.info['mesh'] = str(mesh_dir)
    if skel_dir:
      meta.info['skeletons'] = str(skel_dir)

    image = PrecomputedImageSource(
      config, meta, cache,
      autocrop=bool(autocrop),
      bounded=bool(bounded),
      non_aligned_writes=False,
      fill_missing=bool(fill_missing),
      delete_black_uploads=bool(delete_black_uploads),
      background_color=background_color,
    )

    mesh = GrapheneMeshSource(meta, cache, config)
    skeleton = PrecomputedSkeletonSource(meta, cache, config)

    return CloudVolumeGraphene(
      meta, cache, config, 
      image, mesh, skeleton,
      mip=mip
    )

def register():
  register_plugin('graphene', create_graphene)


# from __future__ import print_function
# from six.moves import range

# import pytest
# import re
# import time

# from cloudvolume.storage import Storage
# from cloudvolume import exceptions, Bbox, chunks
# from layer_harness import delete_layer, TEST_NUMBER


# #TODO delete files created by tests
# def test_read_write():
#   urls = [
#     "file:///tmp/removeme/read_write",
#     "gs://seunglab-test/cloudvolume/read_write",
#     "s3://seunglab-test/cloudvolume/read_write"
#   ]

#   for num_threads in range(0,11,5):
#     for url in urls:
#       url = url + '-' + str(TEST_NUMBER)
#       with Storage(url, n_threads=num_threads) as s:
#         content = b'some_string'
#         s.put_file('info', content, compress=None, cache_control='no-cache')
#         s.wait()
#         assert s.get_file('info') == content
#         assert s.get_file('nonexistentfile') is None

#         num_infos = max(num_threads, 1)

#         results = s.get_files([ 'info' for i in range(num_infos) ])

#         assert len(results) == num_infos
#         assert results[0]['filename'] == 'info'
#         assert results[0]['content'] == content
#         assert all(map(lambda x: x['error'] is None, results))
#         assert s.get_files([ 'nonexistentfile' ])[0]['content'] is None

#         s.delete_file('info')
#         s.wait()

#         s.put_json('info', { 'omg': 'wow' }, cache_control='no-cache')
#         s.wait()
#         results = s.get_json('info')
#         assert results == { 'omg': 'wow' }

#   delete_layer("/tmp/removeme/read_write")

# def test_http_read():
#   with Storage("https://storage.googleapis.com/seunglab-test/test_v0/black/") as stor:
#     info = stor.get_json('info')

#   assert info == {
#     "data_type": "uint8",
#     "num_channels": 1,
#     "scales": [
#       {
#         "chunk_sizes": [
#           [
#             64,
#             64,
#             50
#           ]
#         ],
#         "encoding": "raw",
#         "key": "6_6_30",
#         "resolution": [
#           6,
#           6,
#           30
#         ],
#         "size": [
#           1024,
#           1024,
#           100
#         ],
#         "voxel_offset": [
#           0,
#           0,
#           0
#         ]
#       }
#     ],
#     "type": "image"
#   }


# def test_http_read_brotli_image():
#   fn = "2_2_50/4096-4608_4096-4608_112-128"
#   bbox = Bbox.from_filename(fn) # possible off by one error w/ exclusive bounds

#   with Storage("https://open-neurodata.s3.amazonaws.com/kharris15/apical/em") as stor:
#     img_bytes = stor.get_file(fn)

#   img = chunks.decode(img_bytes, 'raw', shape=bbox.size3(), dtype="uint8")

#   assert img.shape == (512, 512, 16)


# def test_delete():
#   urls = [
#     "file:///tmp/removeme/delete",
#     "gs://seunglab-test/cloudvolume/delete",
#     "s3://seunglab-test/cloudvolume/delete"
#   ]

#   for url in urls:
#     url = url + '-' + str(TEST_NUMBER)
#     with Storage(url, n_threads=1) as s:
#       content = b'some_string'
#       s.put_file('delete-test', content, compress=None, cache_control='no-cache').wait()
#       s.put_file('delete-test-compressed', content, compress='gzip', cache_control='no-cache').wait()
#       assert s.get_file('delete-test') == content
#       s.delete_file('delete-test').wait()
#       assert s.get_file('delete-test') is None

#       assert s.get_file('delete-test-compressed') == content
#       s.delete_file('delete-test-compressed').wait()
#       assert s.get_file('delete-test-compressed') is None

#       # Reset for batch delete
#       s.put_file('delete-test', content, compress=None, cache_control='no-cache').wait()
#       s.put_file('delete-test-compressed', content, compress='gzip', cache_control='no-cache').wait()
#       assert s.get_file('delete-test') == content
#       assert s.get_file('delete-test-compressed') == content

#       s.delete_files(['delete-test', 'delete-nonexistent',
#                       'delete-test-compressed']).wait()
#       assert s.get_file('delete-test') is None
#       assert s.get_file('delete-test-compressed') is None

# def test_compression():
#   urls = [
#     "file:///tmp/removeme/compress",
#     "gs://seunglab-test/cloudvolume/compress",
#     "s3://seunglab-test/cloudvolume/compress"
#   ]

#   compression_tests = [
#     '',
#     None,
#     True,
#     False,
#     'gzip',
#     'br',
#   ]

#   for url in urls:
#     url = url + '-' + str(TEST_NUMBER)
#     for method in compression_tests:
#       with Storage(url, n_threads=5) as s:
#         content = b'some_string'

#         # remove when GCS enables "br"
#         if method == "br" and "gs://" in url:
#           with pytest.raises(
#               exceptions.UnsupportedCompressionType, 
#               match="Brotli unsupported on google cloud storage"
#           ):
#             s.put_file('info', content, compress=method)
#             s.wait()
#             retrieved = s.get_file('info')
#         else:
#           s.put_file('info', content, compress=method)
#           s.wait()
#           retrieved = s.get_file('info')
#           assert content == retrieved

#         assert s.get_file('nonexistentfile') is None

#     with Storage(url, n_threads=5) as s:
#       content = b'some_string'
#       try:
#         s.put_file('info', content, compress='nonexistent').wait()
#         assert False
#       except NotImplementedError:
#         pass

#     if "file://" in url:
#       delete_layer("/tmp/removeme/compress-" + str(TEST_NUMBER))

# @pytest.mark.parametrize("compression_method", ("gzip", "br"))
# def test_compress_level(compression_method):
#   filepath = "/tmp/removeme/compress_level-" + str(TEST_NUMBER)
#   url = "file://" + filepath

#   content = b'some_string' * 1000

#   compress_levels = range(1, 9, 2)
#   for compress_level in compress_levels:
#     with Storage(url, n_threads=5) as s:
#       s.put_file('info', content, compress=compression_method, compress_level=compress_level)
#       s.wait()

#       retrieved = s.get_file('info')
#       assert content == retrieved

#       _, e = s._interface.get_file("info")
#       assert e == compression_method

#       assert s.get_file('nonexistentfile') is None

#     delete_layer(filepath)


# def test_list():  
#   urls = [
#     "file:///tmp/removeme/list",
#     "gs://seunglab-test/cloudvolume/list",
#     "s3://seunglab-test/cloudvolume/list"
#   ]

#   for url in urls:
#     url = url + '-' + str(TEST_NUMBER)
#     with Storage(url, n_threads=5) as s:
#       print('testing service:', url)
#       content = b'some_string'
#       s.put_file('info1', content, compress=None)
#       s.put_file('info2', content, compress=None)
#       s.put_file('build/info3', content, compress=None)
#       s.put_file('level1/level2/info4', content, compress=None)
#       s.put_file('info5', content, compress='gzip')
#       s.put_file('info.txt', content, compress=None)
#       s.wait()
#       time.sleep(1) # sometimes it takes a moment for google to update the list
#       assert set(s.list_files(prefix='')) == set(['build/info3','info1', 'info2', 'level1/level2/info4', 'info5', 'info.txt'])
      
#       assert set(s.list_files(prefix='inf')) == set(['info1','info2','info5','info.txt'])
#       assert set(s.list_files(prefix='info1')) == set(['info1'])
#       assert set(s.list_files(prefix='build')) == set(['build/info3'])
#       assert set(s.list_files(prefix='build/')) == set(['build/info3'])
#       assert set(s.list_files(prefix='level1/')) == set(['level1/level2/info4'])
#       assert set(s.list_files(prefix='nofolder/')) == set([])

#       # Tests (1)
#       assert set(s.list_files(prefix='', flat=True)) == set(['info1','info2','info5','info.txt'])
#       assert set(s.list_files(prefix='inf', flat=True)) == set(['info1','info2','info5','info.txt'])
#       # Tests (2)
#       assert set(s.list_files(prefix='build', flat=True)) == set([])
#       # Tests (3)
#       assert set(s.list_files(prefix='level1/', flat=True)) == set([])
#       assert set(s.list_files(prefix='build/', flat=True)) == set(['build/info3'])
#       # Tests (4)
#       assert set(s.list_files(prefix='build/inf', flat=True)) == set(['build/info3'])

#       for file_path in ('info1', 'info2', 'build/info3', 'level1/level2/info4', 'info5', 'info.txt'):
#         s.delete_file(file_path)
  
#   delete_layer("/tmp/removeme/list")


# def test_exists():
#   urls = [
#     "file:///tmp/removeme/exists",
#     "gs://seunglab-test/cloudvolume/exists",
#     "s3://seunglab-test/cloudvolume/exists"
#   ]

#   for url in urls:
#     url = url + '-' + str(TEST_NUMBER)
#     with Storage(url, n_threads=5) as s:
#       content = b'some_string'
#       s.put_file('info', content, compress=None)
#       s.wait()
#       time.sleep(1) # sometimes it takes a moment for google to update the list
      
#       assert s.exists('info')
#       assert not s.exists('doesntexist')
#       s.delete_file('info')

# def test_access_non_cannonical_paths():
#   urls = [
#     "file:///tmp/noncanon",
#     "gs://seunglab-test/noncanon",
#     "s3://seunglab-test/noncanon"
#   ]

#   for url in urls:
#     url = url + '-' + str(TEST_NUMBER)
#     with Storage(url, n_threads=5) as s:
#       content = b'some_string'
#       s.put_file('info', content, compress=None)
#       s.wait()
#       time.sleep(0.5) # sometimes it takes a moment for google to update the list
      
#       assert s.get_file('info') == content
#       assert s.get_file('nonexistentfile') is None
#       s.delete_file('info')
#       s.wait()


from .image import BossImageSource
from .metadata import BossMetadata

from ...frontends.precomputed import CloudVolumePrecomputed

from ...cacheservice import CacheService
from ...cloudvolume import SharedConfiguration, register_plugin
from ...paths import strict_extract

def create_boss(
    cloudpath, mip=0, bounded=True, autocrop=False,
    fill_missing=False, cache=False, compress_cache=None,
    cdn_cache=True, progress=False, info=None, provenance=None,
    compress=None, non_aligned_writes=False, parallel=1,
    delete_black_uploads=False, green_threads=False
  ):
    path = strict_extract(cloudpath)
    config = SharedConfiguration(
      cdn_cache=cdn_cache,
      compress=compress,
      compress_level=None,
      green=green_threads,
      mip=mip,
      parallel=parallel,
      progress=progress,
    )
    cache = CacheService(
      cloudpath=(cache if type(cache) == str else cloudpath),
      enabled=bool(cache),
      config=config,
      compress=compress_cache,
    )

    meta = BossMetadata(cloudpath, cache=cache, info=info)
    image = BossImageSource(
      config, meta, cache, 
      autocrop=bool(autocrop),
      bounded=bool(bounded),
      non_aligned_writes=bool(non_aligned_writes), 
    )

    return CloudVolumePrecomputed(
      meta, cache, config, 
      imagesrc, mesh=None, skeleton=None,
      mip=mip
    )

def register():
  register_plugin('boss', create_boss)


import os
import pytest
import sys

import numpy as np

from cloudvolume import dask as dasklib

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_3d():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((3, 3, 3), chunks=1)
  with du.tmpdir() as d:
    d = 'file://' + d
    dasklib.to_cloudvolume(a, d)
    a2 = dasklib.from_cloudvolume(d)
    a2 = da.squeeze(a2, axis=3)
    da.utils.assert_eq(a, a2, check_meta=False)  # TODO: add meta check
    assert a.chunks == a2.chunks

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_4d():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((3, 3, 3, 3))
  with du.tmpdir() as d:
    d = 'file://' + d
    dasklib.to_cloudvolume(a, d)
    a2 = dasklib.from_cloudvolume(d)
    da.utils.assert_eq(a, a2, check_meta=False)  # TODO: add meta check
    assert a.chunks == a2.chunks

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_4d_channel_rechunked():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((3, 3, 3, 3), chunks=(2, 2, 2, 3))
  with du.tmpdir() as d:
    d = 'file://' + d
    dasklib.to_cloudvolume(a, d)
    a2 = dasklib.from_cloudvolume(d)
    np.testing.assert_array_equal(a.compute(), a2.compute())
    # Channel has single chunk
    assert a2.chunks == ((2, 1), (2, 1), (2, 1), (3, ))

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_4d_1channel():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((3, 3, 3, 1))
  with du.tmpdir() as d:
    d = 'file://' + d
    dasklib.to_cloudvolume(a, d)
    a2 = dasklib.from_cloudvolume(d)
    da.utils.assert_eq(a, a2, check_meta=False)  # TODO: add meta check
    assert a.chunks == a2.chunks

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_rechunk_3d():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((9, 9, 9), chunks=3)
  with du.tmpdir() as d:
    d = 'file://' + d
    dasklib.to_cloudvolume(a, d)
    a2 = dasklib.from_cloudvolume(d, chunks=5)
    assert a2.chunks == ((5, 4), (5, 4), (5, 4), (1, ))

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_rechunk_4d():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((9, 9, 9, 3), chunks=3)
  with du.tmpdir() as d:
    d = 'file://' + d
    dasklib.to_cloudvolume(a, d)
    a2 = dasklib.from_cloudvolume(d, chunks=5)
    assert a2.chunks == ((5, 4), (5, 4), (5, 4), (3, ))

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_delayed_compute():
  dask = pytest.importorskip('dask')
  da = pytest.importorskip('dask.array')
  dd = pytest.importorskip('dask.delayed')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((3, 3, 3, 1), chunks=1)
  with du.tmpdir() as d:
    d = 'file://' + d
    out = dasklib.to_cloudvolume(a, d, compute=False)
    assert isinstance(out, dd.Delayed)
    dask.compute(out)
    a2 = dasklib.from_cloudvolume(d)
    da.utils.assert_eq(a, a2, check_meta=False)  # TODO: add meta check
    assert a.chunks == a2.chunks

@pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 2 not supported.")
def test_roundtrip_delayed():
  da = pytest.importorskip('dask.array')
  du = pytest.importorskip('dask.utils')
  a = da.zeros((3, 3, 3, 3))
  with du.tmpdir() as d:
    cloudpath = 'file://' + d
    task = dasklib.to_cloudvolume(a, cloudpath, compute=False)
    assert not os.listdir(d)
    task.compute()
    a2 = dasklib.from_cloudvolume(cloudpath)
    da.utils.assert_eq(a, a2, check_meta=False)  # TODO: add meta check
    assert a.chunks == a2.chunks



from collections.abc import Iterable
import json
import os
import posixpath

import json5
import multiprocessing as mp
import numpy as np
from six import string_types
from six.moves import range
from tqdm import tqdm

from cloudfiles import CloudFiles 

from ... import exceptions
from ...provenance import DatasetProvenance, DataLayerProvenance

from ... import lib
from ...lib import ( 
  colorize, red, mkdir, 
  Vec, Bbox, jsonify, 
)
from ...paths import strict_extract, ascloudpath, to_https_protocol

def downscale(size, factor_in_mip, roundingfn):
  smaller = Vec(*size, dtype=np.float32) / Vec(*factor_in_mip)
  return list(map(int, roundingfn(smaller)))

class PrecomputedMetadata(object):
  """
  The PrecomputedMetadataService provides methods for fetching,
  saving, and accessing information about the data type & compression, 
  bounding box, resolution, and provenance of a given dataset 
  stored in Precomputed format.  

  This class is a building block for building a class that can
  read and write Precomputed data types.
  """
  def __init__(
    self, cloudpath, config=None, cache=None, 
    info=None, provenance=None, 
    max_redirects=10, use_https=False
  ):
    self.path = strict_extract(cloudpath)
    self.cache = cache
    if self.cache:
      self.cache.meta = self
    self.config = config
    self.info = None

    self.redirected_from = []
    self.use_https = use_https

    if info is None:
      self.refresh_info(max_redirects=max_redirects)
      if self.cache and self.cache.enabled:
        self.cache.check_info_validity()
    else:
      self.info = info

    if provenance is None:
      self.provenance = None
      self.refresh_provenance()
      if self.cache.enabled:
        self.cache.check_provenance_validity()
    else:
      self.provenance = self._cast_provenance(provenance)

  @classmethod
  def create_info(cls, 
    num_channels, layer_type, data_type, encoding, 
    resolution, voxel_offset, volume_size, 
    mesh=None, skeletons=None, chunk_size=(128,128,64),
    compressed_segmentation_block_size=(8,8,8),
    max_mip=0, factor=Vec(2,2,1), redirect=None
  ):
    """
    Create a new neuroglancer Precomputed info file.

    Required:
      num_channels: (int) 1 for grayscale, 3 for RGB 
      layer_type: (str) typically "image" or "segmentation"
      data_type: (str) e.g. "uint8", "uint16", "uint32", "float32"
      encoding: (str) "raw" for binaries like numpy arrays, "jpeg"
      resolution: int (x,y,z), x,y,z voxel dimensions in nanometers
      voxel_offset: int (x,y,z), beginning of dataset in positive cartesian space
      volume_size: int (x,y,z), extent of dataset in cartesian space from voxel_offset
    
    Optional:
      mesh: (str) name of mesh directory, typically "mesh"
      skeletons: (str) name of skeletons directory, typically "skeletons"
      chunk_size: int (x,y,z), dimensions of each downloadable 3D image chunk in voxels
      compressed_segmentation_block_size: (x,y,z) dimensions of each compressed sub-block
        (only used when encoding is 'compressed_segmentation')
      max_mip: (int), the maximum mip level id.
      factor: (Vec), the downsampling factor for each mip level
      redirect: (str), cloudpath to redirect to

    Returns: dict representing a single mip level that's JSON encodable
    """
    if not isinstance(factor, Vec):
      factor = Vec(*factor)

    if not isinstance(data_type, str):
      data_type = np.dtype(data_type).name

    info = {
      "num_channels": int(num_channels),
      "type": layer_type,
      "data_type": data_type,
      "scales": [{
        "encoding": encoding,
        "chunk_sizes": [chunk_size],
        "key": "_".join(map(str, resolution)),
        "resolution": list(map(int, resolution)),
        "voxel_offset": list(map(int, voxel_offset)),
        "size": list(map(int, volume_size)),
      }],
    }

    if redirect is not None:
      info['redirect'] = str(redirect)
    
    fullres = info['scales'][0]
    factor_in_mip = factor.clone()
 
    # add mip levels
    for _ in range(max_mip):
      new_resolution = list(map(int, Vec(*fullres['resolution']) * factor_in_mip ))
      newscale = {
        u"encoding": encoding,
        u"chunk_sizes": [ list(map(int, chunk_size)) ],
        u"key": "_".join(map(str, new_resolution)),
        u"resolution": new_resolution,
        u"voxel_offset": downscale(fullres['voxel_offset'], factor_in_mip, np.floor),
        u"size": downscale(fullres['size'], factor_in_mip, np.ceil),
      }
      info['scales'].append(newscale)
      factor_in_mip *= factor

    if encoding == 'compressed_segmentation':
      info['scales'][0]['compressed_segmentation_block_size'] = list(map(int, compressed_segmentation_block_size))

    if mesh:
      info['mesh'] = 'mesh' if not isinstance(mesh, string_types) else mesh

    if skeletons:
      info['skeletons'] = 'skeletons' if not isinstance(skeletons, string_types) else skeletons      
    
    return info

  def refresh_info(self, max_redirects=10, force_fetch=False):
    """
    Refresh the current info file from the cache (if enabled) 
    or primary storage (e.g. the cloud) if not cached.

    max_redirects: number of times to allow redirection. set to 0 to
      force getting the origin info file loaded.
    force_fetch: bypass the cache for reading, but allow writing

    Raises:
      cloudvolume.exceptions.InfoUnavailableError when the info file 
        is unable to be retrieved.
      cloudvolume.exceptions.TooManyRedirects if more than max_redirects
        are followed.
      cloudvolume.exceptions.CyclicRedirect if a previously visited 
        location is revisited.

    See also: fetch_info

    Returns: dict
    """
    if self.cache and self.cache.enabled and not force_fetch:
      info = self.cache.get_json('info')
      if info:
        self.info = info
        return self.info

    self.info = self.redirectable_fetch_info(max_redirects)

    if self.cache:
      self.cache.maybe_cache_info()
    return self.info

  def fetch_info(self):
    """
    Refresh the current info file from primary storage (e.g. the cloud) without
    refrence to the cache. The cache will not be updated.
  
    Raises cloudvolume.exceptions.InfoUnavailableError when the info file 
    is unable to be retrieved.

    See also: refresh_info

    Returns: dict
    """
    info = CloudFiles(self.cloudpath, secrets=self.config.secrets).get_json('info')

    if info is None:
      raise exceptions.InfoUnavailableError(
        red('No info file was found: {}'.format(self.infopath))
      )

    return info

  def redirectable_fetch_info(self, max_redirects=10):
    """
    Refresh the current info file from primary storage (e.g. the cloud) without
    refrence to the cache. The cache will not be updated. 'redirect' links
    in the info file will be followed up to `max_redirects` times after which
    an exception will be raised.

    Raises:
      cloudvolume.exceptions.InfoUnavailableError when the info file 
        is unable to be retrieved.
      cloudvolume.exceptions.TooManyRedirects if more than max_redirects
        are followed.
      cloudvolume.exceptions.CyclicRedirect if a previously visited 
        location is revisited.

    See also: refresh_info, fetch_info

    Optional:
      max_redirects: if 'redirect' is specified in an info file, 
        follow the link up to this many times to the pointed locations.

    Returns: dict
    """
    visited = []

    if max_redirects <= 0:
      return self.fetch_info()

    if self.path.format == 'graphene':
      start = self.server_url
    else:
      start = self.cloudpath

    for _ in range(max_redirects):
      info = self.fetch_info()

      if 'redirect' not in info or not info['redirect']:
        break

      path = strict_extract(info['redirect'])
      if self.use_https:
        path = to_https_protocol(path)

      if path == self.path:
        break 
      elif path in visited:
        raise exceptions.CyclicRedirect(
          """
Tried to redirect through a cycle.

Start: {}
Hops: 
\t{}
\n""".format(
          start, 
          "\n\t".join([ 
            str(i+1) + ". " + ascloudpath(v) for i, v in enumerate(visited) 
          ]))
        )

      visited.append(path)
      self.path = path
    else:
      raise exceptions.TooManyRedirects(
        "Tried to redirect more than {} hops.".format(max_redirects)
      )

    self.redirected_from = visited[:-1]

    return info

  def commit_info(self):
    """
    Save the current info dict as JSON into cache and primary storage.

    Raises KeyError if an encoding of 'compressed_segmentation' is specified
    without 'compressed_segmentation_block_size'.

    Raises ValueError if 'compressed_segmentation' is specified and the 
    data type is not uint32 or uint64.
    """
    for scale in self.scales:
      if scale['encoding'] == 'compressed_segmentation':
        if 'compressed_segmentation_block_size' not in scale.keys():
          raise KeyError("""
            'compressed_segmentation_block_size' must be set if 
            compressed_segmentation is set as the encoding.

            A typical value for compressed_segmentation_block_size is (8,8,8)

            Info file specification:
            https://github.com/google/neuroglancer/blob/master/src/neuroglancer/datasource/precomputed/README.md#info-json-file-specification
          """)
        elif self.data_type not in ('uint32', 'uint64'):
          raise ValueError("compressed_segmentation can only be used with uint32 and uint64 data types.")

    infojson = jsonify(self.info, 
      sort_keys=True,
      indent=2, 
      separators=(',', ': ')
    )
    # use put instead of put_json to preserve formatting
    cf = CloudFiles(self.cloudpath, secrets=self.config.secrets)
    cf.put(
      'info', infojson,
      cache_control='no-cache',
      content_type='application/json'
    )

    if self.cache:
      self.cache.maybe_cache_info()

  def refresh_provenance(self):
    """
    Refresh the current irovenance file from the cache (if enabled) 
    or primary storage (e.g. the cloud) if not cached. If the provenance
    file does not exist, no error is raised and None is returned.

    Raises ValueError if the provenance file is present but can not
    be json5 decoded.

    See also: fetch_provenance

    Returns: dict or None
    """
    if self.cache and self.cache.enabled:
      prov = self.cache.get_json('provenance')
      if prov:
        self.provenance = DataLayerProvenance(**prov)
        return self.provenance

    self.provenance = self.fetch_provenance()
    if self.cache:
      self.cache.maybe_cache_provenance()
    return self.provenance

  def _cast_provenance(self, prov):
    if isinstance(prov, DataLayerProvenance):
      return prov
    elif isinstance(prov, string_types):
      prov = json.loads(prov)

    provobj = DataLayerProvenance(**prov)
    provobj.sources = provobj.sources or []  
    provobj.owners = provobj.owners or []
    provobj.processing = provobj.processing or []
    provobj.description = provobj.description or ""
    provobj.validate()
    return provobj

  def fetch_provenance(self):
    """
    Refresh the current provenance file from primary storage (e.g. the cloud)
    without reference to cache. The cache will not be updated.
  
    Raises cloudvolume.exceptions.provenanceUnavailableError when the info file 
    is unable to be retrieved.

    See also: refresh_provenance

    Returns: dict
    """
    cf = CloudFiles(self.cloudpath, secrets=self.config.secrets)
    provfile = cf.get('provenance')
    if provfile:
      provfile = provfile.decode('utf-8')

      # The json5 decoder is *very* slow
      # so use the stricter but much faster json 
      # decoder first, and try it only if it fails.
      try:
        provfile = json.loads(provfile)
      except json.decoder.JSONDecodeError:
        try:
          provfile = json5.loads(provfile)
        except ValueError:
          raise ValueError(red("""The provenance file could not be JSON decoded. 
            Please reformat the provenance file before continuing. 
            Contents: {}""".format(provfile)))
    else:
      provfile = {
        "sources": [],
        "owners": [],
        "processing": [],
        "description": "",
      }

    return self._cast_provenance(provfile)

  def commit_provenance(self):
    """
    Save the current provenance object as JSON into cache (if enabled) 
    and primary storage.
    """
    prov = self.provenance.serialize()

    # hack to pretty print provenance files
    prov = json.loads(prov)
    prov = jsonify(prov, 
      sort_keys=True,
      indent=2, 
      separators=(',', ': ')
    )

    # need to use put vs put_json to preserve formatting
    cf = CloudFiles(self.cloudpath, secrets=self.config.secrets)
    cf.put(
      'provenance', prov, 
      cache_control='no-cache', 
      content_type='application/json'
    )

    if self.cache:
      self.cache.maybe_cache_provenance()

  @property
  def dataset(self):
    return self.path.dataset
  
  @property
  def layer(self):
    return self.path.layer

  @property
  def scales(self):
    return self.info['scales']

  @scales.setter
  def scales(self, val):
    self.info['scales'] = val

  def scale(self, mip):
    return self.info['scales'][mip]

  def join(self, *paths):
    if self.path.protocol == 'file':
      return os.path.join(*paths)
    else:
      return posixpath.join(*paths)

  @property
  def basepath(self):
    return self.path.basepath
    
  @property 
  def layerpath(self):
    return self.join(self.basepath, self.layer)

  @property
  def base_cloudpath(self):
    return self.path.protocol + "://" + self.basepath

  @property 
  def cloudpath(self):
    return self.join(self.base_cloudpath, self.layer)
  
  @property
  def infopath(self):
    return self.join(self.cloudpath, 'info')

  @property
  def skeletons(self):
    return self.info['skeletons'] if 'skeletons' in self.info else None

  @property
  def mesh(self):
    return self.info['mesh'] if 'mesh' in self.info else None

  def shape(self, mip):
    """Returns Vec(x,y,z,channels) shape of the volume similar to numpy.""" 
    size = self.volume_size(mip)
    return Vec(size.x, size.y, size.z, self.num_channels)

  def volume_size(self, mip):
    """Returns Vec(x,y,z) shape of the volume (i.e. shape - channels).""" 
    return Vec(*self.info['scales'][mip]['size'])

  @property
  def available_mips(self):
    """Returns a list of mip levels that are defined."""
    return range(len(self.info['scales']))

  @property
  def available_resolutions(self):
    """Returns a list of defined resolutions."""
    return (s["resolution"] for s in self.scales)

  @property
  def layer_type(self):
    """e.g. 'image' or 'segmentation'"""
    return self.info['type']

  @property
  def dtype(self):
    """e.g. np.uint8"""
    return np.dtype(self.data_type)

  @property
  def data_type(self):
    """e.g. 'uint8'"""
    return self.info['data_type']

  def encoding(self, mip, encoding=None):
    """
    If encoding is provided, set the encoding for this
    mip level. If the encoding is not provided, this is
    a getter.

    Typical values: 'raw', 'jpeg', 'compressed_segmentation'

    Returns encoding for the mip level either way.
    """
    if encoding is None:
      return self.info['scales'][mip]['encoding']

    encoding = encoding.lower()
    scale = self.scale(mip)
    scale['encoding'] = encoding
    if (encoding == 'compressed_segmentation' \
      and 'compressed_segmentation_block_size' not in scale):

      scale['compressed_segmentation_block_size'] = (8,8,8)

    return encoding

  def compressed_segmentation_block_size(self, mip):
    if 'compressed_segmentation_block_size' in self.info['scales'][mip]:
      return self.info['scales'][mip]['compressed_segmentation_block_size']
    return None

  @property
  def num_channels(self):
    return int(self.info['num_channels'])

  def voxel_offset(self, mip):
    """Vec(x,y,z) start of the dataset in voxels"""
    scale = self.scale(mip)
    if 'voxel_offset' in scale:
      return Vec(*scale['voxel_offset'])
    else:
      return Vec(0,0,0)

  def resolution(self, mip):
    """Vec(x,y,z) dimensions of each voxel in nanometers"""
    return Vec(*self.info['scales'][mip]['resolution'])

  def to_mip(self, mip):
    mip = list(mip) if isinstance(mip, Iterable) else int(mip)
    try:
      if isinstance(mip, list):  # mip specified by voxel resolution
        return next(
          (i for (i,s) in enumerate(self.scales) if s["resolution"] == mip)
        )
      else:  # mip specified by index into downsampling hierarchy
        return self.available_mips[mip]
    except Exception:
      if isinstance(mip, list):
        opening_text = "Scale <{}>".format(", ".join(map(str, mip)))
      else:
        opening_text = "MIP {}".format(str(mip))
  
      scales = [ ",".join(map(str, scale)) for scale in self.available_resolutions ]
      scales = [ "<{}>".format(scale) for scale in scales ]
      scales = ", ".join(scales)
      msg = "{} not found. {} available: {}".format(
        opening_text, len(self.available_mips), scales
      )
      raise exceptions.ScaleUnavailableError(msg)

  def downsample_ratio(self, mip):
    """Describes how downsampled the current mip level is as an (x,y,z) factor triple."""
    return self.resolution(mip) / self.resolution(0)

  def chunk_size(self, mip):
    """Underlying chunk size dimensions in voxels. Synonym for underlying."""
    return Vec(*self.scale(mip)['chunk_sizes'][0])

  def key(self, mip):
    """The subdirectory within the data layer containing the chunks for this mip level"""
    return self.scale(mip)['key']

  @property
  def keys(self):
    return [ self.key(mip) for mip in self.available_mips ]

  def bounds(self, mip):
    """Returns a 3D spatial bounding box for the dataset with dimensions in voxels."""
    offset = self.voxel_offset(mip)
    shape = self.volume_size(mip)
    return Bbox( offset, offset + shape )

  def bbox(self, mip):
    bounds = self.bounds(mip)
    minpt = list(bounds.minpt) + [ 0 ]
    maxpt = list(bounds.maxpt) + [ self.num_channels ]
    return Bbox(minpt, maxpt)

  def point_to_mip(self, pt, mip, to_mip):
    pt = Vec(*pt)
    downsample_ratio = self.resolution(mip).astype(np.float32) / self.resolution(to_mip).astype(np.float32)
    return np.floor(pt * downsample_ratio)

  def bbox_to_mip(self, bbox, mip, to_mip):
    """Convert bbox or slices from one mip level to another."""
    bbox = Bbox.create(bbox, self.bounds(mip))

    def one_level(bbox, mip, to_mip):
      original_dtype = bbox.dtype
      # setting type required for Python2
      downsample_ratio = self.resolution(mip).astype(np.float32) / self.resolution(to_mip).astype(np.float32)
      bbox = bbox.astype(np.float64)
      bbox *= downsample_ratio
      bbox.minpt = np.floor(bbox.minpt)
      bbox.maxpt = np.ceil(bbox.maxpt)
      bbox = bbox.astype(original_dtype)
      return bbox

    delta = 1 if to_mip >= mip else -1
    while (mip != to_mip):
      bbox = one_level(bbox, mip, mip + delta)
      mip += delta

    return bbox

  def reset_scales(self):
    """Used for manually resetting downsamples if something messed up."""
    self.info['scales'] = self.info['scales'][0:1]

  def add_resolution(self, res, encodig=None, chunk_size=None, info=None):
    factor = Vec(res, dtype=np.float32) // self.meta.resolution(0)
    return self.add_scale(factor, encoding, chunk_size, info)

  def add_scale(self, factor, encoding=None, chunk_size=None, info=None):
    """
    Generate a new downsample scale to for the info file and return an updated dictionary.
    You'll still need to call self.commit_info() to make it permenant.

    Required:
      factor: int (x,y,z), e.g. (2,2,1) would represent a reduction of 2x in x and y

    Optional:
      encoding: force new layer to e.g. jpeg or compressed_segmentation
      chunk_size: force new layer to new chunk size

    Returns: info dict
    """
    if not info:
      info = self.info

    # e.g. {"encoding": "raw", "chunk_sizes": [[64, 64, 64]], "key": "4_4_40", 
    # "resolution": [4, 4, 40], "voxel_offset": [0, 0, 0], 
    # "size": [2048, 2048, 256]}
    fullres = info['scales'][0]

    # If the voxel_offset is not divisible by the ratio,
    # zooming out will slightly shift the data.
    # Imagine the offset is 10
    #    the mip 1 will have an offset of 5
    #    the mip 2 will have an offset of 2 instead of 2.5 
    #        meaning that it will be half a pixel to the left
    if chunk_size is None:
      chunk_size = lib.find_closest_divisor(fullres['chunk_sizes'][0], closest_to=[64,64,64])

    if encoding is None:
      encoding = fullres['encoding']

    newscale = {
      u"encoding": encoding,
      u"chunk_sizes": [ list(map(int, chunk_size)) ],
      u"resolution": list(map(int, Vec(*fullres['resolution']) * factor )),
      u"voxel_offset": downscale(fullres.get('voxel_offset', (0,0,0)), factor, np.floor),
      u"size": downscale(fullres['size'], factor, np.ceil),
    }

    if newscale['encoding'] == 'compressed_segmentation':
      if 'compressed_segmentation_block_size' in fullres:
        newscale['compressed_segmentation_block_size'] = fullres['compressed_segmentation_block_size']  
      else: 
        newscale['compressed_segmentation_block_size'] = (8,8,8)

    newscale[u'key'] = str("_".join([ str(res) for res in newscale['resolution']]))

    new_res = np.array(newscale['resolution'], dtype=int)

    preexisting = False
    for index, scale in enumerate(info['scales']):
      res = np.array(scale['resolution'], dtype=int)
      if np.array_equal(new_res, res):
        preexisting = True
        info['scales'][index] = newscale
        break

    if not preexisting:    
      info['scales'].append(newscale)

    return newscale

  def lock_mips(self, mips):
    """
    Establishes a write lock on the specified mip levels.
    The lock is written to the cloud info file.
    """
    mips = lib.toiter(mips)
    if max(mips) > max(self.available_mips):
      raise ValueError("Cannot lock a mip level that doesn't exist. Highest mip: {} Got: {}".format(
        max(self.available_mips), mips
      ))

    try:
      self.refresh_info(force_fetch=True)

      for mip in mips:
        self.info['scales'][mip]['locked'] = True

      self.commit_info()
    except Exception as err:
      msg = lib.red("Unable to acquire write lock on mips {}!".format(list(mips)))
      raise exceptions.WriteLockAcquisitionError(msg)

  def unlock_mips(self, mips):
    """
    Releases a write lock on the specified mip levels.
    The lock is written to the cloud info file.
    """
    mips = lib.toiter(mips)
    if max(mips) > max(self.available_mips):
      raise ValueError("Cannot unlock a mip level that doesn't exist. Highest mip: {} Got: {}".format(
        max(self.available_mips), mips
      ))

    try:
      self.refresh_info(force_fetch=True)

      for mip in mips:
        self.info['scales'][mip]['locked'] = False

      self.commit_info()
    except Exception as err:
      msg = lib.yellow("Unable to release lock on mips {}".format(list(mips)))
      raise exceptions.WriteLockReleaseError(msg)

  def locked_mips(self):
    return set([ i for i, scale in enumerate(self.info['scales']) if scale.get('locked', False) ])


from ..sharding import ShardingSpecification, ShardReader
from ....skeleton import Skeleton
from ..spatial_index import CachedSpatialIndex
from ....exceptions import EmptyFileException

class ShardedPrecomputedSkeletonSource(object):
  def __init__(self, meta, cache, config, readonly=False):
    self.meta = meta
    self.cache = cache
    self.config = config
    self.readonly = bool(readonly)

    spec = ShardingSpecification.from_dict(self.meta.info['sharding'])
    self.reader = ShardReader(meta, cache, spec)

    self.spatial_index = None
    if self.meta.spatial_index:
      mip = self.meta.mip or 0
      self.spatial_index = CachedSpatialIndex(
        self.cache, self.config,
        cloudpath=self.meta.layerpath, 
        bounds=self.meta.meta.bounds(mip) * self.meta.meta.resolution(mip),
        chunk_size=self.meta.info['spatial_index']['chunk_size'],
      )

  @property
  def path(self):
    return self.meta.skeleton_path

  def get(self, segids):
    list_return = True
    if type(segids) in (int, float):
      list_return = False
      segids = [ int(segids) ]

    # compress = self.config.compress 
    # if compress is None:
    #   compress = True

    results = []
    binaries = self.reader.get_data(
      segids, self.meta.skeleton_path, 
      progress=self.config.progress
    )

    for segid in segids:
      binary = binaries[segid]
      del binaries[segid]

      if binary is None:
        raise EmptyFileException("segid {} is missing.".format(segid))

      skeleton = Skeleton.from_precomputed(
        binary, segid=segid, 
        vertex_attributes=self.meta.info['vertex_attributes']
      )
      skeleton.transform = self.meta.transform
      results.append(skeleton.physical_space())

    if list_return:
      return results
    else:
      return results[0]

  def upload(self, *args, **kwargs):
    raise NotImplementedError()

  def raw_upload(self, *args, **kwargs):
    raise NotImplementedError()

  def get_bbox(self, bbox):
    if self.spatial_index is None:
      raise IndexError("A spatial index has not been created.")

    segids = self.spatial_index.query(bbox)
    return self.get(segids)


import pytest

import numpy as np

import cloudvolume.lib as lib
from cloudvolume.lib import Bbox, Vec

def test_divisors():

  divisors = (
    (1, (1,)), 
    (2, (1,2)),
    (3, (1,3)),
    (4, (1,2,4)),
    (6, (1,2,3,6)),
    (35, (1,5,7,35)),
    (128, (1,2,4,8,16,32,64,128)),
    (258, (1,2,3,6,43,86,129,258)),
  )

  for num, ans in divisors:
    result = [ _ for _ in lib.divisors(num) ]
    result.sort()
    assert tuple(result) == ans

def test_find_closest_divisor():
  size = lib.find_closest_divisor( (128,128,128), (64,64,64) )
  assert tuple(size) == (64,64,64)

  size = lib.find_closest_divisor( (240,240,240), (64,64,64) )
  assert tuple(size) == (60,60,60)

  size = lib.find_closest_divisor( (224,224,224), (64,64,64) )
  assert tuple(size) == (56,56,56)

  size = lib.find_closest_divisor( (73,73,73), (64,64,64) )
  assert tuple(size) == (73,73,73)

def test_bbox_subvoxel():
  bbox = Bbox( (0,0,0), (1,1,1), dtype=np.float32)
  
  assert not bbox.subvoxel()
  assert not bbox.empty()

  bbox.maxpt[:] *= -1
  bbox.maxpt.z = 20

  # pathological case
  assert not (bbox.volume() < 1)
  assert bbox.subvoxel()
  assert bbox.empty()

  bbox = Bbox( (1,1,1), (1,1,1) )
  assert bbox.empty()

  bbox = Bbox( (0,0,0), (0.9, 1.0, 1.0) )
  assert bbox.subvoxel()
  assert not bbox.empty()

def test_vec_division():
  vec = Vec(2,4,8)
  assert np.all( (vec/2) == Vec(1,2,4) )


def test_bbox_division():
  box = Bbox( (0,2,4), (4,8,16) )
  assert (box//2) == Bbox( (0,1,2), (2,4,8) )

  box = Bbox( (0,3,4), (4,8,16), dtype=np.float32 )
  print((box/2.))
  print(Bbox( (0., 1.5, 2.), (2., 4., 8.) ))
  assert (box/2.) == Bbox( (0., 1.5, 2.), (2., 4., 8.) )

def test_bbox_intersect():
  box = Bbox( (0,0,0), (10, 10, 10) )
  
  assert Bbox.intersects(box, box)
  assert Bbox.intersects(box, Bbox((1,1,1), (11,11,11)) )
  assert Bbox.intersects(box, Bbox((-1,-1,-1), (9,9,9)) ) 
  assert Bbox.intersects(box, Bbox((5, -5, 0), (15, 5, 10)))
  assert not Bbox.intersects(box, Bbox( (30,30,30), (40,40,40) ))
  assert not Bbox.intersects(box, Bbox( (-30,-30,-30), (-40,-40,-40) ))
  assert not Bbox.intersects(box, Bbox( (10, 0, 0), (20, 10, 10) ))

def test_bbox_division():
  bbox = Bbox( (1,1,1), (10, 10, 10), dtype=np.float32 )
  bbox2 = bbox.clone()

  bbox /= 3.0
  bbx3 = bbox2 / 3.0

  point333 = np.float32(1) / np.float32(3)

  assert np.all( np.abs(bbx3.minpt - point333) < 1e-6 )
  assert np.all( bbx3.maxpt == np.float32(3) + point333 )
  assert bbox == bbx3

  bbox = Bbox( (1,1,1), (10, 10, 10), dtype=np.float32 )

  x = bbox.minpt 
  bbox /= 3.0 
  assert np.all(x == point333)

def test_bbox_slicing():
  bbx_rect = Bbox.from_slices(np.s_[1:10,1:10,1:10])
  bbx_plane = Bbox.from_slices(np.s_[1:10,10:1,1:10])

  assert bbx_rect == Bbox((1,1,1), (10,10,10))
  assert bbx_plane == Bbox((1,10,1), (10, 10, 10))

  try:
    bbx_plane = Bbox.from_slices(np.s_[1:10,10:1:-1,1:10])
    assert False
  except ValueError:
    pass

  bbx_plane = Bbox.from_slices(np.s_[1:10,10:1:1,1:10])
  assert bbx_plane == Bbox((1,10,1), (10, 10, 10))


def test_bbox_intersection():
  bbx1 = Bbox( (0,0,0), (10,10,10) )
  bbx2 = Bbox( (5,5,5), (15,15,15) )

  assert Bbox.intersection(bbx1, bbx2) == Bbox((5,5,5),(10,10,10))
  assert Bbox.intersection(bbx2, bbx1) == Bbox((5,5,5),(10,10,10))
  bbx2.minpt = Vec(11,11,11)
  assert Bbox.intersection(bbx1, bbx2) == Bbox((0,0,0),(0,0,0))

def test_bbox_hashing():
  bbx = Bbox.from_list([ 1,2,3,4,5,6 ])
  d = {}
  d[bbx] = 1

  assert len(d) == 1
  for _,v in d.items():
    assert v == 1

  bbx = Bbox( (1., 1.3, 2.), (3., 4., 4.) )
  d = {}
  d[bbx] = 1

  assert len(d) == 1
  for _,v in d.items():
    assert v == 1

def test_bbox_serialize():
  bbx = Bbox( (24.125, 2512.2, 2112.3), (33.,532., 124.12412), dtype=np.float32)

  reconstituted = Bbox.deserialize(bbx.serialize())
  assert bbx == reconstituted

def test_bbox_volume():
  bbx = Bbox( (0,0,0), (2000, 2000, 2000) )
  # important thing is 8B is > int32 size
  assert bbx.volume() == 8000000000

  bbx = bbx.astype(np.float32)
  assert bbx.volume() == 8000000000

def test_jsonify():
  obj = {
    'x': [ np.array([1,2,3,4,5], dtype=np.uint64) ],
    'y': [ {}, {} ],
    'z': np.int32(5),
    'w': '1 2 34 5'
  }

  assert lib.jsonify(obj, sort_keys=True) == r"""{"w": "1 2 34 5", "x": [[1, 2, 3, 4, 5]], "y": [{}, {}], "z": 5}"""


def test_bbox_from_filename():
  filenames = [
    "0-512_0-512_0-16",
    "0-512_0-512_0-16.gz",
    "0-512_0-512_0-16.br",
  ]

  for fn in filenames:
    bbox = Bbox.from_filename(fn)
    assert np.array_equal(bbox.minpt, [0, 0, 0])
    assert np.array_equal(bbox.maxpt, [512, 512, 16])

  filenames = [
    "gibberish",
    "0-512_0-512_0-16.lol",
    "0-512_0-512_0-16.gzip",
    "0-512_0-512_0-16.brotli",
    "0-512_0-512_0-16.na",
    "0-512_0-512_0-abc",
    "0-512_0-abc_0-16",
    "0-abc_0-512_0-16",
  ]
  for fn in filenames:
    with pytest.raises(ValueError):
      bbox = Bbox.from_filename(fn)



"""Utilties for translating to/from dask arrays.

NOTE: Using the thread-based dask scheduler is very inefficient with
      CloudVolume because CV is pure python and will cause substantial
      GIL contention between the dask worker threads.  It is HIGHLY
      ADVISABLE to use a distributed process scheduler with one thread
      per process.

"""
import numpy as np
from .cloudvolume import CloudVolume


def to_cloudvolume(arr,
                   cloudpath,
                   resolution=(1, 1, 1),
                   voxel_offset=(0, 0, 0),
                   layer_type=None,
                   encoding='raw',
                   max_mip=0,
                   compute=True,
                   return_stored=False,
                   **kwargs):
  """Save 3d or 4d dask array to the precomputed CloudVolume storage format.

  NOTE: DO NOT USE thread-based dask scheduler. See comment at top of module.

  See https://docs.dask.org/en/latest/array.html for details about the format.

  Parameters
  ----------
  arr: dask.array
    Data to store
  cloudpath: str
    Path to the dataset layer. This should match storage's supported
    providers.
    e.g. Google: gs://$BUCKET/$DATASET/$LAYER/
         S3    : s3://$BUCKET/$DATASET/$LAYER/
         Lcl FS: file:///tmp/$DATASET/$LAYER/
         Boss  : boss://$COLLECTION/$EXPERIMENT/$CHANNEL
         HTTP/S: http(s)://.../$CHANNEL
         matrix: matrix://$BUCKET/$DATASET/$LAYER/
  resolution: Iterable of ints of length 3
    The x, y, z voxel dimensions in nanometers
  voxel_offset: Iterable of ints of length 3
    The x, y, z beginning of dataset in positive cartesian space.
  layer_type: str
    "image" or "segmentation"
  max_mip: int
    Maximum mip level id.
  compute: boolean, optional
    If true compute immediately, return ``dask.delayed.Delayed`` otherwise.
  return_stored: boolean, optional
    Optionally return stored results.
  kwargs: passed to the ``cloudvolume.CloudVolume()`` function, e.g., compression options

  Raises
  ------
  ValueError
    If ``arr`` has ndim different that 3 or 4, or ``layer_type`` is unsupported.

  Returns
  -------
  See notes on `compute` and `return_stored` parameters.
  """
  import dask
  import dask.array as da
  if not da.core._check_regular_chunks(arr.chunks):
    raise ValueError('Attempt to save array to cloudvolume with irregular '
                     'chunking, please call `arr.rechunk(...)` first.')

  if not layer_type:
    if arr.dtype in (np.bool, np.uint32, np.uint64, np.uint16):
      layer_type = 'segmentation'
    elif np.issubdtype(arr.dtype, np.integer) or np.issubdtype(arr.dtype, np.floating):
      layer_type = 'image'
    else:
      raise ValueError('Unsupported layer_type for CloudVolume: %s' % layer_type)

  if arr.ndim == 3:
    num_channels = 1
    chunk_size = arr.chunksize
  elif arr.ndim == 4:
    num_channels = arr.shape[-1]
    chunk_size = arr.chunksize[:3]
  else:
    raise ValueError('CloudVolume only supports 3 or 4 dimensions.  Array has %d.' % arr.ndim)

  info = CloudVolume.create_new_info(num_channels,
                                     layer_type,
                                     arr.dtype.name,
                                     encoding,
                                     resolution,
                                     voxel_offset,
                                     arr.shape[:3],
                                     chunk_size=chunk_size,
                                     max_mip=max_mip)

  # Delay writing any metadata until computation time.
  #   - the caller may never do the full computation
  #   - the filesystem may be slow, and there is a desire to open files
  #     in parallel on worker machines.
  vol = dask.delayed(_create_cloudvolume)(cloudpath, info, **kwargs)
  return arr.store(vol, lock=False, compute=compute, return_stored=return_stored)


def _create_cloudvolume(cloudpath, info, **kwargs):
  """Create cloudvolume and commit metadata."""
  vol = CloudVolume(cloudpath, info=info, progress=False, **kwargs)
  vol.commit_info()
  vol.provenance.processing = [{'method': 'cloudvolume.dask.to_cloudvolume'}]
  vol.commit_provenance()
  return vol


def from_cloudvolume(cloudpath, chunks=None, name=None, **kwargs):
  """Load dask array from a cloudvolume compatible dataset.

  NOTE: DO NOT USE thread-based dask scheduler. See comment at top of module.

  Volumes with a single channel will be returned as 4d arrays with a
  length-1 channel dimension, even if they were stored from 3d data.

  The channel dimension is returned as a single-chunk by default, as that
  is how CloudVolumes are stored.

  See https://docs.dask.org/en/latest/array.html for details about the format.

  Parameters
  ----------
  cloudpath: str
    Path to the dataset layer. This should match storage's supported
    providers.
    e.g. Google: gs://$BUCKET/$DATASET/$LAYER/
         S3    : s3://$BUCKET/$DATASET/$LAYER/
         Lcl FS: file:///tmp/$DATASET/$LAYER/
         Boss  : boss://$COLLECTION/$EXPERIMENT/$CHANNEL
         HTTP/S: http(s)://.../$CHANNEL
         matrix: matrix://$BUCKET/$DATASET/$LAYER/
  chunks: tuple of ints or tuples of ints
    Passed to ``da.from_array``, allows setting the chunks on
    initialisation, if the chunking scheme in the stored dataset is not
    optimal for the calculations to follow.  Note that the chunking should
    be compatible with an underlying 4d array.
  name: str, optional
     An optional keyname for the array.  Defaults to hashing the input
  kwargs: passed to the ``cloudvolume.CloudVolume()`` function, e.g., compression options

  Returns
  -------
  Dask array
  """
  import dask.array as da
  from dask.base import tokenize

  vol = CloudVolume(cloudpath, progress=False, **kwargs)
  if chunks is None:
      chunks = tuple(vol.chunk_size) + (vol.num_channels, )
  if name is None:
      name = 'from-cloudvolume-' + tokenize(vol, chunks, **kwargs)
  return da.from_array(vol, chunks, name=name)



import os

from .metadata import PrecomputedSkeletonMetadata
from .sharded import ShardedPrecomputedSkeletonSource
from .unsharded import UnshardedPrecomputedSkeletonSource

from ..metadata import PrecomputedMetadata
from ....cacheservice import CacheService
from ....paths import strict_extract
from ....cloudvolume import SharedConfiguration

class PrecomputedSkeletonSource(object):
  def __new__(cls, meta, cache, config, readonly=False):
    skel_meta = PrecomputedSkeletonMetadata(meta, cache)

    if skel_meta.is_sharded():
      return ShardedPrecomputedSkeletonSource(skel_meta, cache, config, readonly) 

    return UnshardedPrecomputedSkeletonSource(skel_meta, cache, config, readonly)

  @classmethod
  def from_cloudpath(cls, cloudpath, cache=False, progress=False):
    config = SharedConfiguration(
      cdn_cache=False,
      compress=True,
      compress_level=None,
      green=False,
      mip=0,
      parallel=1,
      progress=progress,
    )

    cache = CacheService(
      cloudpath=(cache if type(cache) == str else cloudpath),
      enabled=bool(cache),
      config=config,
      compress=True,
    )

    cloudpath, skel_dir = os.path.split(cloudpath)
    meta = PrecomputedMetadata(cloudpath, cache, info={ 'skeletons': skel_dir })

    return PrecomputedSkeletonSource(meta, cache, config)


from six.moves import range
from cloudvolume.threaded_queue import ThreadedQueue
from functools import partial

def test_threading():
  execution_count = 1000
  executions = []


  def reset_executions():
    return [ False for _ in range(execution_count) ]

  def addone(idnum, should_be_none):
    executions[idnum] = True    
    assert should_be_none is None

  executions = reset_executions()

  with ThreadedQueue(n_threads=1) as tq:
    for idnum in range(execution_count):
      fn = partial(addone, idnum)
      tq.put(fn)
  assert all(executions)

  executions = reset_executions()
  tq = ThreadedQueue(n_threads=40)
  for idnum in range(execution_count):
    fn = partial(addone, idnum)
    tq.put(fn)
  tq.wait().kill_threads()
  assert tq.processed == execution_count
  assert all(executions)

  # Base class with 0 threads on with statement will never terminate
  try:
    with ThreadedQueue(n_threads=0) as tq:
      assert False
  except ValueError:
    assert True
  except Exception:
    assert False

def test_derived_class():
  
  def what_fun(should_be_fun):
    assert should_be_fun == 'fun'

  class DerivedThreadedQueue(ThreadedQueue):
    def _initialize_interface(self):
      return 'fun'

  with DerivedThreadedQueue(n_threads=1) as tq:
    for _ in range(1000):
      tq.put(what_fun)

    tq.wait()
    assert tq.processed == 1000

  # shouldn't crash w/ 0 threads because it's a derived class
  with DerivedThreadedQueue(n_threads=0) as tq:
    pass

def test_threads_die():
  tq = ThreadedQueue(n_threads=40)
  assert tq.are_threads_alive()
  tq.kill_threads()
  assert not tq.are_threads_alive()

  tq = ThreadedQueue(n_threads=0)
  assert not tq.are_threads_alive()

  with ThreadedQueue(n_threads=40) as tq:
    threads = tq._threads
  
  assert not any(map(lambda t: t.is_alive(), threads))

def test_thread_exceptions():

  def diediedie(interface):  
    raise NotImplementedError("Not implemented at all.")

  tq = ThreadedQueue(n_threads=40)
  for _ in range(1000):
    tq.put(diediedie)

  try:
    tq.wait()
  except NotImplementedError:
    pass






import re

from ....lib import jsonify

import numpy as np

SKEL_MIP_REGEXP = re.compile(r'skeletons_mip_(\d+)')

class PrecomputedSkeletonMetadata(object):
  def __init__(self, meta, cache=None, info=None):
    self.meta = meta
    self.cache = cache

    if info:
      self.info = info
    elif 'skeletons' in self.meta.info and self.meta.info['skeletons']:
      self.info = self.fetch_info()
    else:
      self.info = self.default_info()

  @property
  def spatial_index(self):
    if 'spatial_index' in self.info:
      return self.info['spatial_index']
    return None  

  @property
  def skeleton_path(self):
    if 'skeletons' in self.meta.info:
      return self.meta.info['skeletons']
    return 'skeletons'

  @property
  def mip(self):
    if 'mip' in self.info:
      return int(self.info['mip'])
    
    # Igneous has long used skeletons_mip_N to store
    # some information about the skeletonizing job. Let's 
    # exploit that for now.
    matches = re.search(SKEL_MIP_REGEXP, self.skeleton_path)
    if matches is None:
      return None 

    mip, = matches.groups()
    return int(mip)

  def join(self, *paths):
    return self.meta.join(*paths)

  @property
  def transform(self):
    return np.array(self.info['transform'], dtype=np.float32).reshape( (3,4) )

  @transform.setter
  def transform(self, val):
    self.info['transform'] = val

  @property
  def basepath(self):
    return self.meta.basepath

  @property
  def cloudpath(self):
    return self.meta.cloudpath

  @property
  def layerpath(self):
    return self.meta.join(self.meta.cloudpath, self.skeleton_path)

  def fetch_info(self):
    info = self.cache.download_json(self.meta.join(self.skeleton_path, 'info'))
    if not info:
      return self.default_info()
    return info

  def refresh_info(self):
    self.info = self.fetch_info()
    return self.info

  def commit_info(self):
    if self.info:
      self.cache.upload_single(
        self.meta.join(self.skeleton_path, 'info'),
        jsonify(self.info), 
        content_type='application/json',
        compress=False,
        cache_control='no-cache',
      )

  def default_info(self):
    return {
      '@type': 'neuroglancer_skeletons',
      'transform': [  
        1, 0, 0, 0, # identity
        0, 1, 0, 0,
        0, 0, 1, 0
      ],
      'vertex_attributes': [
        {
          "id": "radius",
          "data_type": "float32",
          "num_components": 1,
        }, 
        {
          "id": "vertex_types",
          "data_type": "uint8",
          "num_components": 1,
        }
      ],
      'sharding': None,
      'spatial_index': None, # { 'chunk_size': physical units }
    }

  def is_sharded(self):
    if 'sharding' not in self.info:
      return False
    elif self.info['sharding'] is None:
      return False
    else:
      return True



import pytest

import sys

import numpy as np

from cloudvolume.chunks import encode, decode
from cloudvolume import chunks

def encode_decode(data, format, shape=(64,64,64), num_chan=1):
  encoded = encode(data, format)
  result = decode(encoded, format, shape=list(shape) + [ num_chan ], dtype=np.uint8)

  assert np.all(result.shape == data.shape)
  assert np.all(data == result)

def test_kempression():
  data = np.random.random_sample(size=1024 * 3).reshape( (64, 4, 4, 3) ).astype(np.float32)
  encoded = encode(data, 'kempressed')
  result = decode(encoded, 'kempressed', shape=(64, 4, 4, 3), dtype=np.float32)
  assert np.all(result.shape == data.shape)
  assert np.all(np.abs(data - result) <= np.finfo(np.float32).eps)

def test_compressed_segmentation():
  def run_test(shape, block_size, accelerated):
    data = np.random.randint(255, size=shape, dtype=np.uint32)
    encoded = chunks.encode_compressed_segmentation(data, block_size, accelerated)

    compressed = np.frombuffer(encoded, dtype=np.uint32)

    assert compressed[0] == 1 # one channel

    # at least check headers for integrity
    # 64 bit block header 
    # encoded bits (8 bit), lookup table offset (24 bit), encodedValuesOffset (32)
    grid = np.ceil(np.array(shape[3:], dtype=np.float32) / np.array(block_size, dtype=np.float32))
    grid = grid.astype(np.uint32)
    for i in range(np.prod(grid)):
      encodedbits = (compressed[2*i + 1] & 0xff000000) >> 24
      table_offset = compressed[2*i + 1] & 0x00ffffff
      encoded_offset = compressed[2*i + 2]

      assert encodedbits in (0, 1, 2, 4, 8, 16, 32)
      assert table_offset < len(compressed)
      assert encoded_offset < len(compressed)

    result = chunks.decode_compressed_segmentation(encoded, 
      shape=shape,
      dtype=np.uint32,
      block_size=block_size,
      accelerated=accelerated,
    ) 

    assert np.all(data == result)

  try:
    import _compressed_segmentation
    test_options = (True, False)
  except:
    test_options = (False,)

  for accelerated in test_options:
    run_test( ( 2, 2, 2, 1), (2,2,2), accelerated )
    run_test( ( 1, 2, 2, 1), (2,2,2), accelerated )
    run_test( ( 2, 1, 2, 1), (2,2,2), accelerated )
    run_test( ( 2, 2, 1, 1), (2,2,2), accelerated )
    run_test( (64,64,64,1), (8,8,8), accelerated )
    run_test( (16,16,16,1), (8,8,8), accelerated )
    run_test( (8,8,8,1), (8,8,8), accelerated )
    run_test( (4,4,4,1), (8,8,8), accelerated )
    run_test( (4,4,4,1), (2,2,2), accelerated )
    run_test( (2,4,4,1), (2,2,2), accelerated )
  
  if True in test_options:
    run_test( (10,8,8,1), (10,8,8), True ) # known bug in pure python verison

def test_fpzip():
  for N in range(0,100):
    flts = np.array(range(N), dtype=np.float32).reshape( (N,1,1,1) )
    compressed = encode(flts, 'fpzip')
    assert isinstance(compressed, bytes)
    decompressed = decode(compressed, 'fpzip')
    assert np.all(decompressed == flts)

  for N in range(0, 200, 2):
    flts = np.array(range(N), dtype=np.float32).reshape( (N // 2, 2, 1, 1) )
    compressed = encode(flts, 'fpzip')
    assert isinstance(compressed, bytes)
    decompressed = decode(compressed, 'fpzip')
    assert np.all(decompressed == flts)

def test_raw():
  random_data = np.random.randint(255, size=(64,64,64,1), dtype=np.uint8)
  encode_decode(random_data, 'raw')

def test_npz():
  random_data = np.random.randint(255, size=(64,64,64,1), dtype=np.uint8)
  encode_decode(random_data, 'npz')

@pytest.mark.parametrize("shape", ( (64,64,64), (64,61,50)))
@pytest.mark.parametrize("num_channels", (1,3))
def test_jpeg(shape, num_channels):
  xshape = list(shape) + [ num_channels ]
  data = np.zeros(shape=xshape, dtype=np.uint8)
  encode_decode(data, 'jpeg', shape, num_channels)
  encode_decode(data + 255, 'jpeg', shape, num_channels)

  # Random jpeg won't decompress to exactly the same image
  # but it should have nearly the same average power
  random_data = np.random.randint(255, size=xshape, dtype=np.uint8)
  pre_avg = random_data.copy().flatten().mean()
  encoded = encode(random_data, 'jpeg')
  decoded = decode(encoded, 'jpeg', shape=xshape, dtype=np.uint8)
  post_avg = decoded.copy().flatten().mean()

  assert abs(pre_avg - post_avg) < 1




import os

import json
from six.moves import range

import numpy as np
from tqdm import tqdm

from .lib import mkdir, save_images

from . import microviewer

class VolumeCutout(np.ndarray):

  def __new__(cls, buf, path, cloudpath, resolution, mip, layer_type, bounds, handle, *args, **kwargs):
    return super(VolumeCutout, cls).__new__(cls, shape=buf.shape, buffer=np.asfortranarray(buf), dtype=buf.dtype, order='F')

  def __init__(self, buf, path, cloudpath, resolution, mip, layer_type, bounds, handle, *args, **kwargs):
    super(VolumeCutout, self).__init__()

    self.dataset_name = path.dataset
    self.layer = path.layer
    self.path = path
    self.resolution = resolution
    self.cloudpath = cloudpath
    self.mip = mip
    self.layer_type = layer_type
    self.bounds = bounds
    self.handle = handle

  def close(self):
    # This bizzare construction is because of this error:
    # Traceback (most recent call last):
    #   File "cloud-volume/cloudvolume/volumecutout.py", line 30, in __del__
    #     self.close()
    #   File "cloud-volume/cloudvolume/volumecutout.py", line 26, in close
    #     if self.handle and not self.handle.closed:
    # ValueError: mmap closed or invalid

    # However testing if it is closed does not throw an error. So we test
    # for closure and capture the exception if self.handle is None.

    try:
      if not self.handle.closed:
        self.handle.close()
    except AttributeError:
      pass

  # How to add a new attribute to an ndarray:
  # https://docs.scipy.org/doc/numpy-1.13.0/uer/basics.subclassing.html#simple-example-adding-an-extra-attribute-to-ndarray

  # Overriding __getitem__ and __setitem__ are 
  # not easy to do. 
  def __array_finalize__(self, obj):
    if obj is None: 
      return

    self.dataset_name = getattr(obj, 'dataset', None)
    self.layer = getattr(obj, 'layer', None)
    self.path = getattr(obj, 'path', None)
    self.resolution = getattr(obj, 'resolution', None)
    self.cloudpath = getattr(obj, 'cloudpath', None)
    self.mip = getattr(obj, 'mip', None)
    self.layer_type = getattr(obj, 'layer_type', None)
    self.bounds = getattr(obj, 'bounds', None)
    self.handle = None #getattr(obj, 'handle', None)

  def __del__(self):
    sup = super(VolumeCutout, self)
    if hasattr(sup, '__del__'):
      sup.__del__()
    self.close()

  @classmethod
  def from_volume(cls, meta, mip, buf, bounds, handle=None):
    return VolumeCutout(
      buf=buf,
      path=meta.path,
      cloudpath=meta.cloudpath,
      resolution=meta.resolution(mip),
      mip=mip,
      layer_type=meta.layer_type,
      bounds=bounds,
      handle=handle,
    )

  @property
  def num_channels(self):
    return self.shape[3]

  def save_images(
    self, directory=None, axis='z', 
    channel=None, global_norm=True, 
    image_format='PNG', progress=True
  ):
    """See cloudvolume.lib.save_images for more information."""
    if directory is None:
      directory = os.path.join('./saved_images', self.path.dataset, self.path.layer, str(self.mip), self.bounds.to_filename())

    return save_images(
      self, directory, axis, 
      channel, global_norm, image_format, 
      progress
    )

  def viewer(self, port=8080):
    """Start a local web app on the given port that lets you explore this cutout."""
    microviewer.run([ self ], port=port)



from .image import PrecomputedImageSource
from .metadata import PrecomputedMetadata
from .mesh import PrecomputedMeshSource
from .skeleton import PrecomputedSkeletonSource

from ...cloudvolume import register_plugin, SharedConfiguration
from ...cacheservice import CacheService
from ...frontends import CloudVolumePrecomputed
from ...lib import yellow
from ...paths import strict_extract

def create_precomputed(
    cloudpath, mip=0, bounded=True, autocrop=False,
    fill_missing=False, cache=False, compress_cache=None,
    cdn_cache=True, progress=False, info=None, provenance=None,
    compress=None, compress_level=None, non_aligned_writes=False, parallel=1,
    delete_black_uploads=False, background_color=0, 
    green_threads=False, use_https=False,
    max_redirects=10, mesh_dir=None, skel_dir=None,
    secrets=None, **kwargs # absorb graphene arguments
  ):
    path = strict_extract(cloudpath)
    config = SharedConfiguration(
      cdn_cache=cdn_cache,
      compress=compress,
      compress_level=compress_level,
      green=green_threads,
      mip=mip,
      parallel=parallel,
      progress=progress,
      secrets=secrets,
    )

    cache = CacheService(
      cloudpath=(cache if type(cache) == str else cloudpath),
      enabled=bool(cache),
      config=config,
      compress=compress_cache,
    )

    meta = PrecomputedMetadata(
      cloudpath, config=config, cache=cache,
      info=info, provenance=provenance,
      max_redirects=max_redirects,
      use_https=use_https # for parsing redirects
    )

    if mesh_dir:
      meta.info['mesh'] = str(mesh_dir)
    if skel_dir:
      meta.info['skeletons'] = str(skel_dir)

    readonly = bool(meta.redirected_from)

    if readonly:
      print(yellow("""
        Redirected 

        From: {} 
        To:   {}
        Hops: {}

        Volume set to readonly to mitigate accidental
        redirection resulting in writing data to the wrong 
        location.

        To set the data to writable, access the destination
        location directly or set:

        vol.image.readonly = False
        vol.mesh.readonly = False 
        vol.skeleton.readonly = False
      """.format(
        cloudpath, meta.cloudpath, len(meta.redirected_from)
      )))

    image = PrecomputedImageSource(
      config, meta, cache,
      autocrop=bool(autocrop),
      bounded=bool(bounded),
      non_aligned_writes=bool(non_aligned_writes),
      fill_missing=bool(fill_missing),
      delete_black_uploads=bool(delete_black_uploads),
      background_color=background_color,
      readonly=readonly,
    )

    mesh = PrecomputedMeshSource(meta, cache, config, readonly)
    skeleton = PrecomputedSkeletonSource(meta, cache, config, readonly)

    return CloudVolumePrecomputed(
      meta, cache, config, 
      image, mesh, skeleton,
      mip
    )

def register():
  register_plugin('precomputed', create_precomputed)


import os
import re

from cloudvolume import paths
from cloudvolume.paths import strict_extract, extract, ExtractedPath

from cloudvolume.exceptions import UnsupportedProtocolError
from cloudvolume import lib

def test_path_extraction():

  def shoulderror(url):
    try:
        paths.strict_extract(url)
        assert False, url
    except:
        pass

  def okgoogle(url):
    path = paths.extract(url)
    assert path.protocol == 'gs', url
    assert path.bucket == 'bucket', url
    assert path.basepath == 'bucket/dataset', url
    assert path.no_bucket_basepath == 'dataset', url
    assert path.dataset == 'dataset', url
    assert path.layer == 'layer', url

  okgoogle('gs://bucket/dataset/layer') 
  shoulderror('s4://dataset/layer')
  shoulderror('dataset/layer')
  shoulderror('s3://dataset')

  # don't error
  assert (strict_extract('graphene://http://localhost:8080/segmentation/1.0/testvol')
    == ExtractedPath(
      'graphene', 'http', 'localhost:8080', 
      'localhost:8080/segmentation/1.0', 'segmentation/1.0', '1.0', 'testvol'))

  assert (strict_extract('precomputed://gs://fafb-ffn1-1234567/segmentation')
    == ExtractedPath(
      'precomputed', 'gs', 'fafb-ffn1-1234567', 
      'fafb-ffn1-1234567', '', 'fafb-ffn1-1234567', 'segmentation'))

  firstdir = lambda x: '/' + x.split('/')[1]

  homepath = lib.toabs('~')
  homerintermediate = homepath.replace(firstdir(homepath), '')[1:]

  curpath = lib.toabs('.')
  curintermediate = curpath.replace(firstdir(curpath), '')[1:]
  
  match = re.match(r'((?:(?:\w:\\\\)|/).+?)\b', lib.toabs('.'))
  bucket, = match.groups()
  
  print(bucket, curintermediate)

  assert (paths.extract('s3://seunglab-test/intermediate/path/dataset/layer') 
      == ExtractedPath(
        'precomputed', 's3', 'seunglab-test', 
        'seunglab-test/intermediate/path/dataset', 'intermediate/path/dataset', 
        'dataset', 'layer'
      ))

  assert (paths.extract('file:///tmp/dataset/layer') 
      == ExtractedPath(
        'precomputed', 'file', "/tmp", '/tmp/dataset', 'dataset', 'dataset', 'layer'
      ))

  assert (paths.extract('file://seunglab-test/intermediate/path/dataset/layer') 
      == ExtractedPath(
        'precomputed', 'file', firstdir(curpath), 
        os.path.join(bucket, curintermediate, 'seunglab-test/intermediate/path/dataset'),   
        os.path.join(curintermediate, 'seunglab-test', 'intermediate/path/dataset'), 
       'dataset', 'layer'))

  assert (paths.extract('gs://seunglab-test/intermediate/path/dataset/layer') 
      == ExtractedPath(
        'precomputed', 'gs', 'seunglab-test',
        'seunglab-test/intermediate/path/dataset', 'intermediate/path/dataset', 
        'dataset', 'layer'
      ))

  assert (paths.extract('file://~/seunglab-test/intermediate/path/dataset/layer') 
      == ExtractedPath(
        'precomputed', 'file', firstdir(homepath), 
        os.path.join(bucket, homerintermediate, 'seunglab-test', 'intermediate/path/dataset'),
        os.path.join(homerintermediate, 'seunglab-test', 'intermediate/path/dataset'),  
        'dataset', 
        'layer'
      )
  )

  assert (paths.extract('file:///User/me/.cloudvolume/cache/gs/bucket/dataset/layer') 
      == ExtractedPath(
        'precomputed', 'file', '/User', 
        '/User/me/.cloudvolume/cache/gs/bucket/dataset', 
        'me/.cloudvolume/cache/gs/bucket/dataset', 'dataset', 'layer'
      ))

  shoulderror('ou3bouqjsa fkj aojsf oaojf ojsaf')

  okgoogle('gs://bucket/dataset/layer/')
  shoulderror('gs://bucket/dataset/layer/info')

  path = paths.extract('s3://bucketxxxxxx/datasetzzzzz91h8__3/layer1br9bobasjf/')
  assert path.format == 'precomputed'
  assert path.protocol == 's3'
  assert path.bucket == 'bucketxxxxxx'
  assert path.dataset == 'datasetzzzzz91h8__3'
  assert path.layer == 'layer1br9bobasjf'

  path = paths.extract('file:///bucket/dataset/layer/')
  assert path.format == 'precomputed'
  assert path.protocol == 'file'
  assert path.bucket == '/bucket'
  assert path.dataset == 'dataset'
  assert path.layer == 'layer'

  shoulderror('lucifer://bucket/dataset/layer/')
  shoulderror('gs://///')
  shoulderror('gs://seunglab-test//segmentation')

  path = paths.extract('file:///tmp/removeme/layer/')
  assert path.format == 'precomputed'
  assert path.protocol == 'file'
  assert path.bucket == '/tmp'
  assert path.dataset == 'removeme'
  assert path.layer == 'layer'


def test_windows_path_extraction():
  extract = paths.extract(r'file://C:\wow\this\is\a\cool\path', windows=True, disable_toabs=True)
  assert extract.format == 'precomputed'
  assert extract.protocol == 'file'
  assert extract.bucket == 'C:\\wow'
  assert extract.basepath == 'C:\\wow\\this\\is\\a\\cool'
  assert extract.no_bucket_basepath == 'this\\is\\a\\cool'  
  assert extract.dataset == 'cool'  
  assert extract.layer == 'path'  

  extract = paths.extract('file://C:\\wow\\this\\is\\a\\cool\\path\\', windows=True, disable_toabs=True)
  assert extract.format == 'precomputed'
  assert extract.protocol == 'file'
  assert extract.bucket == 'C:\\wow'
  assert extract.basepath == 'C:\\wow\\this\\is\\a\\cool'
  assert extract.no_bucket_basepath == 'this\\is\\a\\cool'  
  assert extract.dataset == 'cool'  
  assert extract.layer == 'path'  

  extract = paths.extract('precomputed://https://storage.googleapis.com/neuroglancer-public-data/kasthuri2011/ground_truth', windows=True)
  assert extract.format == 'precomputed'
  assert extract.protocol == 'https'
  assert extract.bucket == 'storage.googleapis.com'
  assert extract.basepath == 'storage.googleapis.com/neuroglancer-public-data/kasthuri2011'
  assert extract.no_bucket_basepath == 'neuroglancer-public-data/kasthuri2011'  
  assert extract.dataset == 'kasthuri2011'  
  assert extract.layer == 'ground_truth'  



from __future__ import print_function
from six.moves import range

import tenacity
from tqdm import tqdm

from cloudvolume.connectionpools import S3ConnectionPool, GCloudBucketPool
from cloudvolume.threaded_queue import ThreadedQueue
from cloudvolume.storage import Storage

S3_POOL = S3ConnectionPool('s3', 'seunglab-test')
GC_POOL = GCloudBucketPool('seunglab-test')

retry = tenacity.retry(
    reraise=True, 
    stop=tenacity.stop_after_attempt(7), 
    wait=tenacity.wait_full_jitter(0.5, 60.0),
)

def test_gc_stresstest():
  with Storage('gs://seunglab-test/cloudvolume/connection_pool/', n_threads=0) as stor:
    stor.put_file('test', 'some string')

  n_trials = 500
  pbar = tqdm(total=n_trials)

  @retry
  def create_conn(interface):
    # assert GC_POOL.total_connections() <= GC_POOL.max_connections * 5
    bucket = GC_POOL.get_connection()
    blob = bucket.get_blob('cloudvolume/connection_pool/test')
    blob.download_as_string()
    GC_POOL.release_connection(bucket)
    pbar.update()

  with ThreadedQueue(n_threads=20) as tq:
    for _ in range(n_trials):
      tq.put(create_conn)

  pbar.close()

def test_s3_stresstest():
  with Storage('s3://seunglab-test/cloudvolume/connection_pool/', n_threads=0) as stor:
    stor.put_file('test', 'some string')

  n_trials = 500
  pbar = tqdm(total=n_trials)

  @retry
  def create_conn(interface):
    conn = S3_POOL.get_connection()  
    # assert S3_POOL.total_connections() <= S3_POOL.max_connections * 5
    bucket = conn.get_object(
      Bucket='seunglab-test',
      Key='cloudvolume/connection_pool/test',
    )
    S3_POOL.release_connection(conn)
    pbar.update()

  with ThreadedQueue(n_threads=20) as tq:
    for _ in range(n_trials):
      tq.put(create_conn)

  pbar.close()



from functools import partial

import tempfile
import cloudvolume
import numpy as np
import shutil
import posixpath
import pytest
import os
from scipy import sparse 
import sys
import json
import re
tempdir = tempfile.mkdtemp()
TEST_PATH = "file://{}".format(tempdir)
TEST_DATASET_NAME = "testvol"
PRECOMPUTED_MESH_TEST_DATASET_NAME = "meshvol_precompute"
DRACO_MESH_TEST_DATASET_NAME = "meshvol_draco"
GRAPHENE_SHARDED_MESH_TEST_DATASET_NAME = "meshvol_graphene_sharded"
PCG_LOCATION = "https://www.dynamicannotationframework.com/"
TEST_SEG_ID = 144115188084020434
TEST_GRAPHENE_SHARDED_ID = 864691135213153080
TEST_TOKEN = "2371270ab23f129cc121dedbeef01294"

def test_graphene_auth_token(graphene_vol):
  cloudpath = "graphene://" + posixpath.join(PCG_LOCATION, 'segmentation', 'api/v1/', TEST_DATASET_NAME)
  
  cloudvolume.CloudVolume(cloudpath, secrets=TEST_TOKEN)
  cloudvolume.CloudVolume(cloudpath, secrets={ "token": TEST_TOKEN })

  try:
    cloudvolume.CloudVolume(cloudpath, secrets=None)
  except cloudvolume.exceptions.AuthenticationError:
    pass

  try:
    cloudvolume.CloudVolume(cloudpath, secrets={ "token": "Z@(ASINAFSOFAFOSNS" })
    assert False
  except cloudvolume.exceptions.AuthenticationError:
    pass

@pytest.fixture()
def cv_graphene_mesh_precomputed(requests_mock):
  test_dir = os.path.dirname(os.path.abspath(__file__))
  graphene_test_cv_dir = os.path.join(test_dir,'test_cv')
  graphene_test_cv_path = "file://{}".format(graphene_test_cv_dir)

  info_d = {
    "app": {
      "supported_api_versions": [
        0,
        1
      ]
    },
    "data_dir": graphene_test_cv_path,
    "data_type": "uint64",
    "graph": {
      "chunk_size": [
        512,
        512,
        128
      ],
      "n_layers": 9,
      "n_bits_for_layer_id": 8,
      "spatial_bit_masks": {
        x: 10 for x in range(255)
      },
    },
    "chunks_start_at_voxel_offset": False,
    "mesh": "mesh_mip_2_err_40_sv16",
    "num_channels": 1,
    "scales": [{
      "chunk_sizes": [[ 512, 512, 16 ]],
      "compressed_segmentation_block_size": [
        8,
        8,
        8
      ],
      "encoding": "compressed_segmentation",
      "key": "8_8_40",
      "resolution": [
        8,
        8,
        40
      ],
      "size": [
        43520,
        26112,
        2176
      ],
      "voxel_offset": [
        17920,
        14848,
        0
      ]
    }],
    "type": "segmentation"
  }
  requests_mock.get(posixpath.join(PCG_LOCATION,
                    'segmentation/table',
                    PRECOMPUTED_MESH_TEST_DATASET_NAME,
                    "info"), json=info_d)
  
  frag_files = os.listdir(os.path.join(graphene_test_cv_dir, info_d['mesh']))
  # the file are saved as .gz but we want to list the non gz version
  # as cloudvolume will take care of finding the compressed files
  frag_files = [f[:-3] for f in frag_files if f[0]=='9']
  frag_d = {'fragments':frag_files}
  mock_url = posixpath.join(PCG_LOCATION,
              "meshing/api/v1/table",
              PRECOMPUTED_MESH_TEST_DATASET_NAME,
              "manifest/{}:0?verify=True".format(TEST_SEG_ID))
  requests_mock.get(mock_url, json=frag_d)

  cloudpath =   posixpath.join(PCG_LOCATION,
                             'segmentation/table',
                             PRECOMPUTED_MESH_TEST_DATASET_NAME)
  yield cloudvolume.CloudVolume("graphene://" + cloudpath, secrets=TEST_TOKEN)

@pytest.fixture()
def cv_graphene_mesh_draco(requests_mock):
  test_dir = os.path.dirname(os.path.abspath(__file__))
  graphene_test_cv_dir = os.path.join(test_dir,'test_cv')
  graphene_test_cv_path = "file://{}".format(graphene_test_cv_dir)

  info_d = {
      "app": {
      "supported_api_versions": [
        0,
        1
      ]
    },
    "data_dir": graphene_test_cv_path,
    "data_type": "uint64",
    "graph": {
      "chunk_size": [
        512,
        512,
        128
      ],
      "n_layers": 9,
      "n_bits_for_layer_id": 8,
      "spatial_bit_masks": {
        x: 10 for x in range(255)
      },
    },
    "chunks_start_at_voxel_offset": False,
    "mesh": "mesh_mip_2_draco_sv16",
    "mesh_metadata": {
      "max_meshed_layer": 6,
      "uniform_draco_grid_size": 21
    },
    "num_channels": 1,
    "scales": [{
      "chunk_sizes": [[ 512, 512, 16 ]],
      "compressed_segmentation_block_size": [
        8,
        8,
        8
      ],
      "encoding": "compressed_segmentation",
      "key": "8_8_40",
      "resolution": [
        8,
        8,
        40
      ],
      "size": [
        43520,
        26112,
        2176
      ],
      "voxel_offset": [
        17920,
        14848,
        0
      ]
    }],
    "type": "segmentation"
  }
  infourl = posixpath.join(PCG_LOCATION,
                           'segmentation/table',
                           DRACO_MESH_TEST_DATASET_NAME,
                           "info")

  requests_mock.get(infourl, json=info_d)
  
  frag_files = os.listdir(os.path.join(graphene_test_cv_dir, info_d['mesh']))
  # we want to filter out the manifest file
  frag_files = [ f for f in frag_files if f[0] == '1' ]
  frag_d = { 'fragments': frag_files }
  mock_url = posixpath.join(
    PCG_LOCATION, 'meshing/api/v1/table', 
    DRACO_MESH_TEST_DATASET_NAME, 
    "manifest/{}:0?verify=True".format(TEST_SEG_ID)
  )
  requests_mock.get(mock_url, json=frag_d)
  cloudpath = posixpath.join(PCG_LOCATION,
                             'segmentation/table',
                             DRACO_MESH_TEST_DATASET_NAME)
  yield cloudvolume.CloudVolume('graphene://' + cloudpath, secrets=TEST_TOKEN)


@pytest.fixture()
def cv_graphene_sharded(requests_mock):
  test_dir = os.path.dirname(os.path.abspath(__file__))
  graphene_test_cv_dir = os.path.join(test_dir,'test_cv')
  graphene_test_cv_path = "gs://seunglab-test/graphene/meshes"

  with open(os.path.join(graphene_test_cv_dir, 'sharded_info.json'), 'r') as fp:
    info_d = json.load(fp)
  info_d['data_dir']=graphene_test_cv_path
  
  infourl = posixpath.join(PCG_LOCATION,
                           'segmentation/table',
                           GRAPHENE_SHARDED_MESH_TEST_DATASET_NAME,
                           "info")
  requests_mock.get(infourl, json=info_d)
  
  valid_manifest={
  "fragments": [
    "~6/29568-0.shard:765877565:4454",
    "~6/50112-0.shard:129695820:17794",
    "~7/3296-0.shard:727627771:13559",
    "~7/3264-2.shard:660015424:21225",
    "~7/6400-3.shard:478017968:31760",
    "~7/7424-2.shard:9298231:40730",
    "~7/4320-0.shard:13324264:53780",
    "~6/29568-0.shard:27890566:21061",
    "516154738544909386:0:40960-49152_57344-65536_0-16384"
  ],
  "seg_ids": [
    440473154180453586,
    446120245900606131,
    511651138917622208,
    511580770173215172,
    518476907102810232,
    520728706916532712,
    513902938730988744,
    440473154180397181,
    516154738544909386
  ]
}
  speculative_manifest = {
  "fragments": [
    "~440473154180453586:6:440473154180087808:29568-0.shard:929",
    "~511651138917622208:7:511651138915794944:3296-0.shard:481",
    "~520728706916532712:7:520728706914713600:7424-2.shard:824",
    "~518476907102810232:7:518476907101028352:6400-3.shard:699",
    "~511580770173215172:7:511580770171617280:3264-2.shard:745",
    "~513902938730988744:7:513902938729480192:4320-0.shard:170",
    "~446120245900606131:6:446120245900345344:50112-0.shard:629",
    "~440473154180397181:6:440473154180087808:29568-0.shard:38",
    "516154738544909386:0:40960-49152_57344-65536_0-16384"
  ],
  "seg_ids": [
    440473154180453586,
    511651138917622208,
    520728706916532712,
    518476907102810232,
    511580770173215172,
    513902938730988744,
    446120245900606131,
    440473154180397181,
    516154738544909386
  ]
}
  verify_manifest_url = posixpath.join(
    PCG_LOCATION, 'meshing/api/v1/table', 
    GRAPHENE_SHARDED_MESH_TEST_DATASET_NAME, 
    "manifest/{}:0?verify=True".format(TEST_GRAPHENE_SHARDED_ID)
  )
  speculative_manifest_url = posixpath.join(
    PCG_LOCATION, 'meshing/api/v1/table', 
    GRAPHENE_SHARDED_MESH_TEST_DATASET_NAME, 
    "manifest/{}:0?verify=False".format(TEST_GRAPHENE_SHARDED_ID)
  )

  requests_mock.get(verify_manifest_url, json=valid_manifest)
  requests_mock.get(speculative_manifest_url, json=speculative_manifest)
  matcher = re.compile('https://storage.googleapis.com/')

  requests_mock.get(matcher,real_http=True)
  cloudpath = posixpath.join(PCG_LOCATION, 'segmentation/table/', GRAPHENE_SHARDED_MESH_TEST_DATASET_NAME)
  yield cloudvolume.CloudVolume('graphene://' + cloudpath, use_https=True, secrets=TEST_TOKEN)


@pytest.fixture(scope='session')
def cv_supervoxels(N=64, blockN=16):

  block_per_row = int(N / blockN)

  chunk_size = [32, 32, 32]
  info = cloudvolume.CloudVolume.create_new_info(
    num_channels=1,
    layer_type='segmentation',
    data_type='uint64',
    encoding='raw',
    resolution=[4, 4, 40],  # Voxel scaling, units are in nanometers
    voxel_offset=[0, 0, 0],  # x,y,z offset in voxels from the origin
    # Pick a convenient size for your underlying chunk representation
    # Powers of two are recommended, doesn't need to cover image exactly
    chunk_size=chunk_size,  # units are voxels
    volume_size=[N, N, N],
  )

  vol = cloudvolume.CloudVolume(TEST_PATH, info=info)
  vol.commit_info()
  xx, yy, zz = np.meshgrid(*[np.arange(0, N) for cs in chunk_size])
  id_ind = (
    np.uint64(xx / blockN),
    np.uint64(yy / blockN),
    np.uint64(zz / blockN)
  )
  id_shape = (block_per_row, block_per_row, block_per_row)

  seg = np.ravel_multi_index(id_ind, id_shape)
  vol[:] = np.uint64(seg)

  yield TEST_PATH

  shutil.rmtree(tempdir)


@pytest.fixture()
def graphene_vol(cv_supervoxels,  requests_mock, monkeypatch, N=64):

  chunk_size = [32, 32, 32]

  info_d = {
    "data_dir": cv_supervoxels,
    "data_type": "uint64",
    "graph": {
      "chunk_size": [64, 64, 64]
    },
    "mesh": "mesh_mip_2_err_40_sv16",
    "num_channels": 1,
    "scales": [
      {
        "chunk_sizes": [
          [32, 32, 32]
        ],
        "compressed_segmentation_block_size": [8, 8, 8],
        "encoding": "compressed_segmentation",
        "key": "4_4_40",
        "resolution": [4, 4, 40],
        "size": [N, N, N],
        "voxel_offset": [0, 0, 0]
      }
    ],
    "type": "segmentation"
  }

  infourl = posixpath.join(PCG_LOCATION, 'segmentation/table', TEST_DATASET_NAME, "info")
  requests_mock.get(infourl, json=info_d)
  
  def mock_get_leaves(self, root_id, bbox, mip):
    return np.array([0,1,2,3], dtype=np.uint64)

  cloudpath = "graphene://" + posixpath.join(PCG_LOCATION, 'segmentation', 'api/v1/', TEST_DATASET_NAME)

  gcv = cloudvolume.CloudVolume(cloudpath, secrets=TEST_TOKEN)
  gcv.get_leaves = partial(mock_get_leaves, gcv)
  yield gcv

def test_gcv(graphene_vol):
  cutout = graphene_vol.download(np.s_[0:5,0:5,0:5], segids=[999])
  assert (np.all(cutout==999))
  cutout_sv = graphene_vol[0:5,0:5,0:5]
  assert cutout_sv.shape == (5,5,5,1)
  assert graphene_vol[0,0,0].shape == (1,1,1,1)


def faces_to_edges(faces, return_index=False):
  """
  Given a list of faces (n,3), return a list of edges (n*3,2)
  Parameters
  -----------
  faces : (n, 3) int
    Vertex indices representing faces
  Returns
  -----------
  edges : (n*3, 2) int
    Vertex indices representing edges
  """
  faces = np.asanyarray(faces)

  # each face has three edges
  edges = faces[:, [0, 1, 1, 2, 2, 0]].reshape((-1, 2))

  if return_index:
    # edges are in order of faces due to reshape
    face_index = np.tile(np.arange(len(faces)),
               (3, 1)).T.reshape(-1)
    return edges, face_index
  return edges

  
def create_csgraph(vertices, edges, euclidean_weight=True, directed=False):
  '''
  Builds a csr graph from vertices and edges, with optional control
  over weights as boolean or based on Euclidean distance.
  '''
  if euclidean_weight:
    xs = vertices[edges[:,0]]
    ys = vertices[edges[:,1]]
    weights = np.linalg.norm(xs-ys, axis=1)
    use_dtype = np.float32
  else:   
    weights = np.ones((len(edges),)).astype(np.int8)
    use_dtype = np.int8 

  if directed:
    edges = edges.T
  else:
    edges = np.concatenate([edges.T, edges.T[[1, 0]]], axis=1)
    weights = np.concatenate([weights, weights]).astype(dtype=use_dtype)

  csgraph = sparse.csr_matrix((weights, edges),
                shape=[len(vertices), ] * 2,
                dtype=use_dtype)

  return csgraph

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_decode_segid(cv_graphene_mesh_draco):
  decoded = cv_graphene_mesh_draco.meta.decode_label(648518346349515986)
  assert decoded.level == 9
  assert decoded.x == 0
  assert decoded.y == 0
  assert decoded.z == 0
  assert decoded.segid == 8164562

  level = '00000101' # 5
  x = '0000000001' # 1 
  y = '0000000010' # 2 
  z = '0000000011' # 3
  segid = '00000000000000000000001010' # 10

  label = int(level + x + y + z + segid, 2)
  decoded = cv_graphene_mesh_draco.meta.decode_label(label)
  
  assert decoded.level == 5
  assert decoded.x == 1
  assert decoded.y == 2
  assert decoded.z == 3
  assert decoded.segid == 10

  encoded = cv_graphene_mesh_draco.meta.encode_label(*decoded)
  assert decoded == cv_graphene_mesh_draco.meta.decode_label(encoded)


@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_graphene_mesh_get_precomputed(cv_graphene_mesh_precomputed):

  mesh = cv_graphene_mesh_precomputed.mesh.get(TEST_SEG_ID)
  edges = faces_to_edges(mesh[TEST_SEG_ID].faces)
  graph = create_csgraph(mesh[TEST_SEG_ID].vertices,
               edges,
               directed=False)
  ccs, labels =  sparse.csgraph.connected_components(graph,
                             directed=False)
  assert(ccs==3)

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_graphene_mesh_get_draco(cv_graphene_mesh_draco):

  mesh = cv_graphene_mesh_draco.mesh.get(TEST_SEG_ID)
  edges = faces_to_edges(mesh[TEST_SEG_ID].faces)
  graph = create_csgraph(mesh[TEST_SEG_ID].vertices,
               edges,
               directed=False)
  ccs, labels =  sparse.csgraph.connected_components(graph,
                             directed=False)
  assert(ccs==3)

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_graphene_mesh_get_graphene_sharded(cv_graphene_sharded):

  mesh = cv_graphene_sharded.mesh.get(TEST_GRAPHENE_SHARDED_ID)
  edges = faces_to_edges(mesh[TEST_GRAPHENE_SHARDED_ID].faces)
  graph = create_csgraph(mesh[TEST_GRAPHENE_SHARDED_ID].vertices,
               edges,
               directed=False)
  ccs, labels =  sparse.csgraph.connected_components(graph,
                             directed=False)
  assert(ccs==21)





from __future__ import print_function

from collections import namedtuple, defaultdict
import copy
import json
from operator import itemgetter
from os.path import basename
import struct

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from . import mmh3
from ... import compression
from ...lib import jsonify, toiter, first
from ...lru import LRU
from ...exceptions import SpecViolation, EmptyFileException

ShardLocation = namedtuple('ShardLocation', 
  ('shard_number', 'minishard_number', 'remainder')
)

uint64 = np.uint64

class ShardingSpecification(object):
  def __init__(
    self, type, preshift_bits, 
    hash, minishard_bits, 
    shard_bits, 
    minishard_index_encoding='raw', 
    data_encoding='raw'
  ):

    self.type = type 
    self.preshift_bits = uint64(preshift_bits)
    self.hash = hash 
    self.minishard_bits = uint64(minishard_bits)
    self.shard_bits = uint64(shard_bits)
    self.minishard_index_encoding = minishard_index_encoding
    self.data_encoding = data_encoding

    self.minishard_mask = self.compute_minishard_mask(self.minishard_bits)
    self.shard_mask = self.compute_shard_mask(self.shard_bits, self.minishard_bits)              

    self.validate()

  def clone(self):
    return ShardingSpecification.from_dict(self.to_dict())

  def index_length(self):
    return int((2 ** self.minishard_bits) * 16)

  @property
  def hash(self):
    return self._hash

  @hash.setter
  def hash(self, val):
    if val == 'identity':
      self.hashfn = lambda x: uint64(x)
    elif val == 'murmurhash3_x86_128':
      self.hashfn = lambda x: uint64(mmh3.hash64(uint64(x).tobytes(), x64arch=False)[0]) 
    else:
      raise SpecViolation("hash {} must be either 'identity' or 'murmurhash3_x86_128'".format(val))

    self._hash = val

  @property
  def preshift_bits(self):
    return self._preshift_bits
  
  @preshift_bits.setter
  def preshift_bits(self, val):
    self._preshift_bits = uint64(val) 

  @property
  def shard_bits(self):
    return self._shard_bits
  
  @shard_bits.setter
  def shard_bits(self, val):
    self._shard_bits = uint64(val) 

  @property
  def minishard_bits(self):
    return self._minishard_bits
  
  @minishard_bits.setter
  def minishard_bits(self, val):
    val = uint64(val)
    self.minishard_mask = self.compute_minishard_mask(val)
    self._minishard_bits = uint64(val)

  def compute_minishard_mask(self, val):
    if val < 0:
      raise ValueError(str(val) + " must be greater or equal to than zero.")
    elif val == 0:
      return uint64(0)

    minishard_mask = uint64(1)
    for i in range(val - uint64(1)):
      minishard_mask <<= uint64(1)
      minishard_mask |= uint64(1)
    return uint64(minishard_mask)

  def compute_shard_mask(self, shard_bits, minishard_bits):
    ones64 = uint64(0xffffffffffffffff)
    movement = uint64(minishard_bits + shard_bits)
    shard_mask = ~((ones64 >> movement) << movement)
    minishard_mask = self.compute_minishard_mask(minishard_bits)
    return shard_mask & (~minishard_mask)

  @classmethod
  def from_json(cls, vals):
    dct = json.loads(vals.decode('utf8'))
    return cls.from_dict(dct)

  def to_json(self):
    return jsonify(self.to_dict())

  @classmethod
  def from_dict(cls, vals):
    vals = copy.deepcopy(vals)
    vals['type'] = vals['@type']
    del vals['@type']
    return cls(**vals)

  def to_dict(self):
    return {
      '@type': self.type,
      'preshift_bits': self.preshift_bits,
      'hash': self.hash,
      'minishard_bits': self.minishard_bits,
      'shard_bits': self.shard_bits,
      'minishard_index_encoding': self.minishard_index_encoding,
      'data_encoding': self.data_encoding,
    }

  def compute_shard_location(self, key):
    chunkid = uint64(key) >> uint64(self.preshift_bits)
    chunkid = self.hashfn(chunkid)
    minishard_number = uint64(chunkid & self.minishard_mask)
    shard_number = uint64((chunkid & self.shard_mask) >> uint64(self.minishard_bits))
    shard_number = format(shard_number, 'x').zfill(int(np.ceil(self.shard_bits / 4.0)))
    remainder = chunkid >> uint64(self.minishard_bits + self.shard_bits)

    return ShardLocation(shard_number, minishard_number, remainder)

  def synthesize_shards(self, data, progress=False):
    """
    Given this specification and a comprehensive listing of
    all the items that could be combined into a given shard,
    synthesize the shard files for this set of labels.

    data: { label: binary, ... }

    e.g. { 5: b'...', 7: b'...' }

    Returns: {
      $filename: binary data,
    }
    """
    return synthesize_shard_files(self, data, progress)

  def synthesize_shard(self, labels, progress=False, presorted=False):
    """
    Assemble a shard file from a group of labels that all belong in the same shard.

    Assembles the .shard file like:
    [ shard index; minishards; all minishard indices ]

    label_group: 
      If presorted is True:
        { minishardno: { label: binary, ... }, ... }
      If presorted is False:
        { label: binary }
    progress: show progress bars

    Returns: binary representing a shard file 
    """
    return synthesize_shard_file(self, labels, progress, presorted)

  def validate(self):
    if self.type not in ('neuroglancer_uint64_sharded_v1',):
      raise SpecViolation(
        "@type ({}) must be 'neuroglancer_uint64_sharded_v1'." \
        .format(self.type)
      )

    if not (64 > self.preshift_bits >= 0):
      raise SpecViolation("preshift_bits must be a whole number less than 64: {}".format(self.preshift_bits))

    if not (64 >= self.minishard_bits >= 0):
      raise SpecViolation("minishard_bits must be between 0 and 64 inclusive: {}".format(self.minishard_bits))

    if not (64 >= self.shard_bits >= 0):
      raise SpecViolation("shard_bits must be between 0 and 64 inclusive: {}".format(self.shard_bits))

    if self.minishard_bits + self.shard_bits > 64:
      raise SpecViolation(
        "minishard_bits and shard_bits must sum to less than or equal to 64: minishard_bits<{}> + shard_bits<{}> = {}".format(
        self.minishard_bits, self.shard_bits, self.minishard_bits + self.shard_bits
      ))

    if self.hash not in ('identity', 'murmurhash3_x86_128'):
      raise SpecViolation("hash {} must be either 'identity' or 'murmurhash3_x86_128'".format(self.hash))

    if self.minishard_index_encoding not in ('raw', 'gzip'):
      raise SpecViolation("minishard_index_encoding only supports values 'raw' or 'gzip'.")

    if self.data_encoding not in ('raw', 'gzip'):
      raise SpecViolation("data_encoding only supports values 'raw' or 'gzip'.")
    
  def __str__(self):
    return "ShardingSpecification::" + str(self.to_dict())

class ShardReader(object):
  def __init__(
    self, meta, cache, spec,
    shard_index_cache_size=512,
    minishard_index_cache_size=128,
    green=False
  ):
    """
    Reads standard Precomputed shard files. 

    meta: a PrecomputedMetadata class
    cache: a CacheService instance
    spec: a ShardingSpecification instance

    shard_index_cache_size: size of LRU cache for fixed indices 
    minishard_index_cache_size: size of LRU cache for minishard indices
    """
    self.meta = meta
    self.cache = cache
    self.spec = spec
    self.green = green

    self.shard_index_cache = LRU(shard_index_cache_size)
    self.minishard_index_cache = LRU(minishard_index_cache_size)

  def get_filename(self, label):
    return self.compute_shard_location(label)[0]

  def compute_shard_location(self, label):
    """
    Returns (filename, shard_number) for meshes and skeletons. 
    Images require a different scheme.
    """
    shard_loc = self.spec.compute_shard_location(label)
    filename = str(shard_loc.shard_number) + '.shard'
    return (filename, shard_loc.minishard_number)

  def get_index(self, filename, path=""):
    """
    Retrieves the shard index which is used for 
    locating the appropriate minishard index.

    Returns: 2^minishard_bits entries of a uint64 
      array of [[ byte start, byte end ], ... ] 
    """
    indices = self.get_indices([ filename ], path, progress=False)
    index = list(indices.values())[0]
    if index is None:
      raise EmptyFileException(filename + " was zero bytes.")
    return index

  def get_indices(self, filenames, path="", progress=None):
    """
    For all given files, retrieves the shard index which 
    is used for locating the appropriate minishard indices.

    Returns: { 
      path_to_/filename.shard: 2^minishard_bits entries of a uint64 
            array of [[ byte start, byte end ], ... ],
      ...
    } 
    """
    filenames = toiter(filenames)
    filenames = [ self.meta.join(path, fname) for fname in filenames ]
    fufilled = { 
      fname: self.shard_index_cache[fname] \
      for fname in filenames \
      if fname in self.shard_index_cache  
    }

    requests = []
    for fname in filenames:
      if fname in fufilled:
        continue
      requests.append({
        'path': fname, 
        'local_alias': fname + '.index', 
        'start': 0, 
        'end': self.spec.index_length(),
      })

    progress = 'Shard Indices' if progress else False
    binaries = self.cache.download_as(requests, progress=progress)
    for (fname, start, end), content in binaries.items():
      try:
        index = self.decode_index(content, fname)
        self.shard_index_cache[fname] = index
        fufilled[fname] = index
      except EmptyFileException:
        self.shard_index_cache[fname] = None
        fufilled[fname] = None

    return fufilled

  def decode_index(self, binary, filename='Shard'):
    if binary is None or len(binary) == 0:
      raise EmptyFileException(filename + " was zero bytes.")
    elif len(binary) != self.spec.index_length():
      raise SpecViolation(
        filename + ": shard index was an incorrect length ({}) for this specification ({}).".format(
          len(binary), self.spec.index_length()
        ))
    
    index = np.frombuffer(binary, dtype=np.uint64)
    index = index.reshape( (index.size // 2, 2), order='C' )
    return index + self.spec.index_length()

  def decode_minishard_index(self, minishard_index, filename=''):
    """Returns [[label, offset, size], ... ] where offset and size are in bytes."""

    if self.spec.minishard_index_encoding != 'raw':
      minishard_index = compression.decompress(
        minishard_index, encoding=self.spec.minishard_index_encoding, filename=filename
      )

    minishard_index = np.copy(np.frombuffer(minishard_index, dtype=np.uint64))
    minishard_index = minishard_index.reshape( (3, len(minishard_index) // 3), order='C' ).T

    for i in range(1, minishard_index.shape[0]):
      minishard_index[i, 0] += minishard_index[i-1, 0]
      minishard_index[i, 1] += minishard_index[i-1, 1] + minishard_index[i-1, 2]

    minishard_index[:,1] += self.spec.index_length()

    return minishard_index 

  def get_minishard_index(self, filename, index, minishard_no, path=""):
    """
    Retrieves the minishard index for a given minishard number.

    Returns: uint64 Nx3 array with multiple rows of [segid, byte start, byte end]
    """
    res = self.get_minishard_indices(filename, index, minishard_no, path)
    return res[minishard_no]

  def get_minishard_indices(self, filename, index, minishard_nos, path=""):
    """
    Retrieves the minishard indices for a set of minishard numbers.

    Returns: { minishard_no: uint64 Nx3 array of [segid, byte start, byte end], ... }
    """
    res = self.get_minishard_indices_for_files(( (filename, index, minishard_nos), ), path)
    return res[basename(filename)]

  def get_minishard_indices_for_files(self, requests, path="", progress=None):
    """
    Fetches the specified minishard indices for all the specified files
    at once. This is required to get high performance as opposed to fetching
    the all minishard indices for a single file.

    requests: iterable of tuples
      [  (filename, index, minishard_numbers), ... ]

    Returns: map of filename -> minishard numbers -> minishard indices

    e.g. 
    {
      filename_1: {
          0: uint64 Nx3 array of [segid, byte start, byte end],
          1: ...,
      }
      filename_2: ...
    }
    """
    fufilled_by_filename = defaultdict(dict)
    msn_map = {}

    download_requests = []
    for filename, index, minishard_nos in requests:
      fufilled_requests, pending_requests = self.compute_minishard_index_requests(
        filename, index, minishard_nos, path
      ) 
      fufilled_by_filename[filename] = fufilled_requests
      for msn, start, end in pending_requests:
        msn_map[(basename(filename), start, end)] = msn

        filepath = self.meta.join(path, filename)

        download_requests.append({
          'path': filepath,
          'local_alias': '{}-{}.msi'.format(filepath, msn),
          'start': start,
          'end': end,
        })

    progress = 'Minishard Indices' if progress else False
    results = self.cache.download_as(download_requests, progress=progress)
  
    for (filename, start, end), content in results.items():
      filename = basename(filename)
      cache_key = (filename, start, end)
      msn = msn_map[cache_key]
      minishard_index = self.decode_minishard_index(content, filename)
      self.minishard_index_cache[cache_key] = minishard_index
      fufilled_by_filename[filename][msn] = minishard_index

    return fufilled_by_filename

  def compute_minishard_index_requests(self, filename, index, minishard_nos, path=""):
    """
    Helper method for get_minishard_indices_for_files. 
    Computes which requests must be made over the network vs can be fufilled from LRU cache.
    """
    minishard_nos = toiter(minishard_nos)

    if index is None:
      return ({ msn: None for msn in minishard_nos }, [])

    fufilled_requests = {}

    byte_ranges = {}
    for msn in minishard_nos:
      bytes_start, bytes_end = index[msn]

      # most typically: [0,0] for an incomplete shard
      if bytes_start == bytes_end:
        fufilled_requests[msn] = None
        continue

      bytes_start, bytes_end = int(bytes_start), int(bytes_end)
      byte_ranges[msn] = (bytes_start, bytes_end)

    full_path = self.meta.join(self.meta.cloudpath, path)

    pending_requests = []
    for msn, (bytes_start, bytes_end) in byte_ranges.items():
      cache_key = (filename, bytes_start, bytes_end)
      if cache_key in self.minishard_index_cache:
        fufilled_requests[msn] = self.minishard_index_cache[cache_key]
      else:
        pending_requests.append((msn, bytes_start, bytes_end))

    return (fufilled_requests, pending_requests)

  def exists(self, labels, path="", return_byte_range=False, progress=None):
    """
    Checks a shard's minishard index for whether a file exists.

    If return_byte_range = False:
      OUTPUT = SHARD_FILEPATH or None if not exists
    Else:
      OUTPUT = [ SHARD_FILEPATH or None, byte_start, num_bytes ]

    Returns:
      If labels is not an iterable:
        return OUTPUT
      Else:
        return { label_1: OUTPUT, label_2: OUTPUT, ... }
    """
    return_one = False

    try:
      iter(labels)
    except TypeError:
      return_one = True

    to_labels = defaultdict(list)
    to_all_labels = defaultdict(list)
    filename_to_minishard_num = defaultdict(list)

    for label in set(toiter(labels)):
      filename, minishard_number = self.compute_shard_location(label)
      to_labels[(filename, minishard_number)].append(label)
      to_all_labels[filename].append(label)
      filename_to_minishard_num[filename].append(minishard_number)

    indices = self.get_indices(to_all_labels.keys(), path, progress=progress)

    all_minishards = self.get_minishard_indices_for_files([ 
      (basename(filepath), index, filename_to_minishard_num[basename(filepath)]) \
      for filepath, index in indices.items()
    ], path, progress=progress)

    results = {}
    for filename, file_minishards in all_minishards.items():
      filepath = self.meta.join(path, filename)
      for mini_no, msi in file_minishards.items():
        labels = to_labels[(filename, mini_no)]

        for label in labels:
          if msi is None:
            results[label] = None
            continue

          idx = np.where(msi[:,0] == label)[0]
          if len(idx) == 0:
            results[label] = None
          else:
            if return_byte_range:
              _, offset, size = msi[idx,:][0]
              results[label] = [ filepath, int(offset), int(size) ]
            else:
              results[label] = filepath

    if return_one:
      return(list(results.values())[0])
    return results

  def disassemble_shard(self, shard):
    """
    Given an entire shard as a bytestring, convert 
    it into a dict of { label: byte content }.
    """
    index = self.decode_index(shard[:self.spec.index_length()])
    shattered = {}
    for start, end in index:
      start, end = int(start), int(end)
      if start == end:
        continue

      msi = self.decode_minishard_index(shard[start:end])
      for label, offset, size in msi:
        offset, size = int(offset), int(size)
        binary = shard[offset:offset+size]
        
        if self.spec.data_encoding != 'raw':
          binary = compression.decompress(binary, encoding=self.spec.data_encoding)
        
        shattered[label] = binary

    return shattered

  def get_data(
    self, label, path="", 
    progress=None, parallel=1
  ):
    """Fetches data from shards.

    label: one or more segment ids
    path: subdirectory path
    progress: display progress bars
    parallel: (int >= 0) use multiple processes

    Return: 
      if label is a scalar:
        a byte string
      else: (label is an iterable)
        {
          label_1: byte string,
          ....
        }
    """
    label, return_multiple = toiter(label, is_iter=True)
    label = set(( int(l) for l in label))
    if not label:
      return {}

    cached = {}
    if self.cache.enabled:
      cached = self.cache.get([ 
        self.meta.join(path, str(lbl)) for lbl in label
      ], progress=progress)

    results = {}
    for cloudpath, content in cached.items():
      lbl = int(basename(cloudpath))
      if content is not None:
        label.remove(lbl)
        
      results[lbl] = cached[cloudpath]

    del cached

    # { label: [ filename, byte start, num_bytes ] }
    exists = self.exists(label, path, return_byte_range=True, progress=progress)
    for k in list(exists.keys()):
      if exists[k] is None:
        results[k] = None
        del exists[k]

    key_label = { (basename(v[0]), v[1], v[2]): k for k,v in exists.items() }

    files = ( 
      { 'path': basename(ext[0]), 'start': int(ext[1]), 'end': int(ext[1]) + int(ext[2]) }
      for ext in exists.values()
    )

    # Requesting many individual shard chunks is slow, but due to z-ordering
    # we might be able to combine adjacent byte ranges. Especially helpful
    # when downloading entire shards!
    bundles = []
    for chunk in sorted(files, key=itemgetter("path", "start")):
      if not bundles or (chunk['path'] != bundles[-1]['path']) or (chunk['start'] != bundles[-1]['end']):
        bundles.append(dict(content=None, subranges=[], **chunk))
      else:
        bundles[-1]['end'] = chunk['end']

      bundles[-1]['subranges'].append({
          'start': chunk['start'],
          'length': chunk['end'] - chunk['start'],
          'slices': slice(chunk['start'] - bundles[-1]['start'], chunk['end'] - bundles[-1]['start'])
      })

    full_path = self.meta.join(self.meta.cloudpath, path)
    bundles_resp = CloudFiles(
      full_path, 
      progress=progress, 
      green=self.green,
      parallel=parallel,
    ).get(bundles)

    # Responses are not guaranteed to be in order of requests
    bundles_resp = { (r['path'], r['byte_range']): r for r in bundles_resp }

    binaries = {}
    for bundle_req in bundles:
      bundle_resp = bundles_resp[(bundle_req['path'], (bundle_req['start'], bundle_req['end']))]
      if bundle_resp['error']:
        raise bundle_resp['error']

      for chunk in bundle_req['subranges']:
        key = (bundle_req['path'], chunk['start'], chunk['length'])
        lbl = key_label[key]
        binaries[lbl] = bundle_resp['content'][chunk['slices']]

    del bundles
    del bundles_resp

    if self.spec.data_encoding != 'raw':
      for filepath, binary in tqdm(binaries.items(), desc="Decompressing", disable=(not progress)):
        if binary is None:
          continue
        binaries[filepath] = compression.decompress(
          binary, encoding=self.spec.data_encoding, filename=filepath
        )
    
    if self.cache.enabled:
      self.cache.put([ 
        (self.meta.join(path, str(filepath)), binary) for filepath, binary in binaries.items()
      ], progress=progress)

    results.update(binaries)

    if return_multiple:
      return results
    return first(results.values())

  def list_labels(self, filename, path="", size=False):
    """
    List all the labels in the index of a given shard file.

    size: (bool) if True, list the size in bytes of each label

    Returns: 
      if not size:
        np.uint64 array of labels 
      else:
        [ (label, size in bytes), ... ] in descending order of size
    """
    index = self.get_index(filename, path)
    all_minishard_nos = list(range(len(index)))
    minishard_indices = self.get_minishard_indices(filename, index, all_minishard_nos, path)
    minishard_indices = [  
      msi for msi in minishard_indices.values() if msi is not None
    ]
    if not size:
      labels = np.concatenate([  
        msi[:,0] for msi in minishard_indices
      ])
      return np.sort(labels)
    else:
      labels = np.concatenate([  
        msi[:,:]
        for msi in minishard_indices
      ])
      labels = [ (row[0], row[2]) for row in labels[:] ]
      return sorted(labels, key=lambda x: x[1], reverse=True)

def synthesize_shard_files(spec, data, progress=False):
  """
  From a set of data guaranteed to constitute one or more
  complete and comprehensive shards (no partial shards) 
  return a set of files ready for upload.

  WARNING: This function is only appropriate for Precomputed
  meshes and skeletons. Use the synthesize_shard_file (singular)
  function to create arbitrarily named and assigned shard files.

  spec: a ShardingSpecification
  data: { label: binary, ... }

  Returns: { filename: binary, ... }
  """
  shard_groupings = defaultdict(lambda: defaultdict(dict))
  pbar = tqdm(
    data.items(), 
    desc='Creating Shard Groupings', 
    disable=(not progress)
  )

  for label, binary in pbar:
    loc = spec.compute_shard_location(label)
    shard_groupings[loc.shard_number][loc.minishard_number][label] = binary

  shard_files = {}

  pbar = tqdm(
    shard_groupings.items(), 
    desc="Synthesizing Shard Files", 
    disable=(not progress)
  )

  for shardno, shardgrp in pbar:
    filename = str(shardno) + '.shard'
    shard_files[filename] = synthesize_shard_file(spec, shardgrp, progress=(progress > 1), presorted=True)

  return shard_files

# NB: This is going to be memory hungry and can be optimized
def synthesize_shard_file(spec, label_group, progress=False, presorted=False):
  """
  Assemble a shard file from a group of labels that all belong in the same shard.

  Assembles the .shard file like:
  [ shard index; minishards; all minishard indices ]

  spec: ShardingSpecification
  label_group: 
    If presorted is True:
      { minishardno: { label: binary, ... }, ... }
    If presorted is False:
      { label: binary }
  progress: show progress bars

  Returns: binary representing a shard file
  """
  minishardnos = []
  minishard_indicies = []
  minishards = []

  if presorted:
    minishard_mapping = label_group
  else:
    minishard_mapping = defaultdict(dict)
    pbar = tqdm(label_group.items(), disable=(not progress), desc="Assigning Minishards")
    for label, binary in pbar:
      loc = spec.compute_shard_location(label)
      minishard_mapping[loc.minishard_number][label] = binary

  del label_group

  for minishardno, minishardgrp in tqdm(minishard_mapping.items(), desc="Minishard Indices", disable=(not progress)):
    labels = sorted([ int(label) for label in minishardgrp.keys() ])
    if len(labels) == 0:
      continue

    minishard_index = np.zeros( (3, len(labels)), dtype=np.uint64, order='C')
    minishard_components = []
    # label and offset are delta encoded
    last_label = 0
    for i, label in enumerate(labels):
      binary = minishardgrp[label]
      if spec.data_encoding != 'raw':
        binary = compression.compress(binary, method=spec.data_encoding)

      minishard_index[0, i] = label - last_label
      minishard_index[1, i] = 0 # minishard_index[2, i - 1]
      minishard_index[2, i] = len(binary)
      minishard_components.append(binary)
      last_label = label
      del minishardgrp[label]

    minishard = b"".join(minishard_components)
    minishardnos.append(minishardno)
    minishard_indicies.append(minishard_index) 
    minishards.append(minishard)

  del minishard_mapping

  cum_minishard_size = 0
  for idx, minishard in zip(minishard_indicies, minishards):
    idx[1, 0] = cum_minishard_size
    cum_minishard_size += len(minishard)

  if progress:
    print("Partial assembly of minishard indicies and data... ", end="", flush=True)

  variable_index_part = [ idx.tobytes('C') for idx in minishard_indicies ]
  if spec.minishard_index_encoding != 'raw':
    variable_index_part = [ 
      compression.compress(idx, method=spec.minishard_index_encoding) \
      for idx in variable_index_part 
    ]

  data_part = b''.join(minishards)
  del minishards

  if progress:
    print("Assembled.")

  fixed_index = np.zeros( 
    (int(2 ** spec.minishard_bits), 2), 
    dtype=np.uint64, order='C'
  )

  start = len(data_part)
  end = len(data_part)
  for i, idx in zip(minishardnos, variable_index_part):
    start = end
    end += len(idx)
    fixed_index[i, 0] = start
    fixed_index[i, 1] = end

  if progress:
    print("Final assembly... ", end="", flush=True)

  # The order here is important. The fixed index must go first because the locations
  # of the other parts are calculated with it implicitly in front. The variable
  # index must go last because otherwise compressing it will affect offset of the
  # data it is attempting to index.

  result = fixed_index.tobytes('C') + data_part + b''.join(variable_index_part) 

  if progress:
    print("Done.")

  return result



import datetime
import os

import posixpath

import numpy as np

from cloudvolume import lib
from cloudvolume.exceptions import (
  SkeletonDecodeError, SkeletonEncodeError, 
  SkeletonUnassignedEdgeError
)
from cloudvolume.lib import red, Bbox

from ..common import cdn_cache_control
from ..spatial_index import CachedSpatialIndex
from ... import readonlyguard

from ....skeleton import Skeleton

class UnshardedPrecomputedSkeletonSource(object):
  def __init__(self, meta, cache, config, readonly=False):
    self.meta = meta
    self.cache = cache
    self.config = config

    self.readonly = bool(readonly)

    self.spatial_index = None
    if self.meta.spatial_index:
      mip = self.meta.mip or 0
      self.spatial_index = CachedSpatialIndex(
        self.cache, self.config,
        cloudpath=self.meta.layerpath, 
        bounds=self.meta.meta.bounds(mip) * self.meta.meta.resolution(mip),
        chunk_size=self.meta.info['spatial_index']['chunk_size'],
      )

  @property
  def path(self):
    return self.meta.skeleton_path 

  def get(self, segids, allow_missing=False):
    """
    Retrieve one or more skeletons from the data layer.

    Example: 
      skel = vol.skeleton.get(5)
      skels = vol.skeleton.get([1, 2, 3])

    Raises SkeletonDecodeError on missing files or decoding errors.

    Required:
      segids: list of integers or integer
    Optional:
      allow_missing: skip over non-existent files otherwise
        raise cloudvolume.exceptions.SkeletonDecodeError

    Returns: 
      if segids is a list, returns list of Skeletons
      else returns a single Skeleton
    """
    list_return = True
    if type(segids) in (int, float):
      list_return = False
      segids = [ int(segids) ]

    compress = self.config.compress 
    if compress is None:
      compress = True

    results = self.cache.download(
      [ self.meta.join(self.meta.skeleton_path, str(segid)) for segid in segids ],
      compress=compress
    )
    missing = [ filename for filename, content in results.items() if content is None ]
    results = { filename: content for filename, content in results.items() if content is not None }

    if not allow_missing and len(missing):
      raise SkeletonDecodeError("File(s) do not exist: {}".format(", ".join(missing)))

    skeletons = []
    for filename, content in results.items():
      segid = int(os.path.basename(filename))
      try:
        skel = Skeleton.from_precomputed(content, segid=segid)
      except Exception as err:
        raise SkeletonDecodeError("segid " + str(segid) + ": " + str(err))
      skel.transform = self.meta.transform
      skeletons.append(skel.physical_space())

    if list_return:
      return skeletons

    if len(skeletons):
      return skeletons[0]

    return None

  @readonlyguard
  def upload_raw(self, segid, vertices, edges, radii=None, vertex_types=None):
    skel = Skeleton(
      vertices, edges, radii, 
      vertex_types, segid=segid
    )
    return self.upload(skel)
    
  @readonlyguard
  def upload(self, skeletons):
    if type(skeletons) == Skeleton:
      skeletons = [ skeletons ]

    files = [ (self.meta.join(self.meta.skeleton_path, str(skel.id)), skel.to_precomputed()) for skel in skeletons ]
    self.cache.upload(
      files=files, 
      compress='gzip', 
      cache_control=cdn_cache_control(self.config.cdn_cache)
    )

  def get_bbox(self, bbox):
    if self.spatial_index is None:
      raise IndexError("A spatial index has not been created.")

    segids = self.spatial_index.query(bbox)
    return self.get(segids)



import re

from ....lib import jsonify

import numpy as np

MESH_MIP_REGEXP = re.compile(r'mesh_mip_(\d+)')

class PrecomputedMeshMetadata(object):
  def __init__(self, meta, cache=None, info=None):
    self.meta = meta
    self.cache = cache

    if info:
      self.info = info
    else:
      self.info = self.fetch_info()

  @property
  def chunk_size(self):
    if 'chunk_size' in self.info:
      return self.info['chunk_size']
    return None

  @property
  def mip(self):
    if 'mip' in self.info:
      return int(self.info['mip'])
    
    # Igneous has long used mesh_mip_N_err_M to store
    # some information about the meshing job. Let's 
    # exploit that for now.
    matches = re.search(MESH_MIP_REGEXP, self.mesh_path)
    if matches is None:
      return None 

    mip, = matches.groups()
    return int(mip)
  
  @property
  def spatial_index(self):
    if 'spatial_index' in self.info:
      return self.info['spatial_index']
    return None

  @property
  def mesh_path(self):
    if 'mesh' in self.meta.info:
      return self.meta.info['mesh']
    return 'mesh'

  def join(self, *paths):
    return self.meta.join(*paths)

  @property
  def basepath(self):
    return self.meta.basepath

  @property
  def cloudpath(self):
    return self.meta.cloudpath

  @property
  def layerpath(self):
    return self.meta.join(self.meta.cloudpath, self.mesh_path)

  def fetch_info(self):
    if 'mesh' not in self.meta.info or not self.meta.info['mesh']:
      return self.default_info()

    info = self.cache.download_json(self.meta.join(self.mesh_path, 'info'))
    if not info:
      return self.default_info()
    return info

  def refresh_info(self):
    self.info = self.fetch_info()
    return self.info

  def commit_info(self):
    if self.info:
      self.cache.upload_single(
        self.meta.join(self.mesh_path, 'info'),
        jsonify(self.info), 
        content_type='application/json',
        compress=False,
        cache_control='no-cache',
      )

  def default_info(self):
    return {
      '@type': 'neuroglancer_legacy_mesh',
      'spatial_index': None, # { 'chunk_size': physical units }
    }

  def is_sharded(self):
    if 'sharding' not in self.info:
      return False
    elif self.info['sharding'] is None:
      return False
    else:
      return True

  def is_multires(self):
    return self.info['@type'] == 'neuroglancer_multilod_draco'

  def is_legacy(self):
    return not self.is_multires()